# -*- coding: utf-8 -*-
# Copyright 2016-2025 The pyXem developers
#
# This file is part of pyXem.
#
# pyXem is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyXem is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyXem.  If not, see <http://www.gnu.org/licenses/>.

"""Utils for Clustering."""

import numpy as np
from sklearn import cluster
from hyperspy.misc.utils import isiterable
import hyperspy.api as hs


def _find_nearest(array, value):
    array = np.asarray(array)
    idx = (np.abs(array - value)).argmin()
    return array[idx]


def _find_max_indices_4D_peak_array(peak_array):
    """Find maximum indices in a 4D peak array.

    Parameters
    ----------
    peak_array : 4D NumPy array

    Returns
    -------
    max_indices : tuple
        (max_x_index, max_y_index)

    Examples
    --------
    >>> import pyxem.utils.cluster_tools as ct
    >>> peak_array0 = np.random.randint(10, 255, size=(3, 4, 5000, 1))
    >>> peak_array1 = np.random.randint(5, 127, size=(3, 4, 5000, 1))
    >>> peak_array = np.concatenate((peak_array0, peak_array1), axis=3)
    >>> max_x_index, max_y_index = ct._find_max_indices_4D_peak_array(
    ...     peak_array)

    """
    max_x_index, max_y_index = 0, 0
    for ix, iy in np.ndindex(peak_array.shape[:2]):
        temp_peak_array = peak_array[ix, iy]
        x_max = int(temp_peak_array[:, 0].max())
        y_max = int(temp_peak_array[:, 1].max())
        if x_max > max_x_index:
            max_x_index = x_max
        if y_max > max_y_index:
            max_y_index = y_max
    return max_x_index, max_y_index


def _filter_4D_peak_array(
    peak_array, signal_axes=None, max_x_index=255, max_y_index=255
):
    """Remove false positives at the outer edges.

    Parameters
    ----------
    peak_array : NumPy array
    signal_axes : HyperSpy signal axes axes_manager, optional
    max_x_index, max_y_index : scalar, optional
        Default 255.

    Examples
    --------
    >>> import pyxem.utils.cluster_tools as ct
    >>> peak_array = np.random.randint(0, 255, size=(3, 4, 100, 2))
    >>> peak_array_filtered = ct._filter_4D_peak_array(peak_array)

    See Also
    --------
    _filter_peak_array_radius
    _filter_peak_list
    _filter_peak_list_radius

    """
    if signal_axes is not None:
        max_x_index = signal_axes[0].high_index
        max_y_index = signal_axes[1].high_index
    peak_array_shape = _get_peak_array_shape(peak_array)
    peak_array_filtered = np.empty(shape=peak_array_shape, dtype=object)
    for index in np.ndindex(peak_array_shape):
        islice = np.s_[index]
        peak_list_filtered = _filter_peak_list(
            peak_array[islice], max_x_index=max_x_index, max_y_index=max_y_index
        )
        peak_array_filtered[islice] = np.array(peak_list_filtered)
    return peak_array_filtered


def _filter_peak_list(peak_list, max_x_index=255, max_y_index=255):
    """Remove false positive peaks at the outer edges.

    Parameters
    ----------
    peak_list : 2D NumPy or 2D list
        [[x0, y0], [x1, y1], ...]
    max_x_index, max_y_index : int, optional
        Default 255.

    Returns
    -------
    peak_list_filtered : 2D NumPy array or 2D list

    Examples
    --------
    >>> import pyxem.utils.cluster_tools as ct
    >>> peak_list = [[128, 129], [10, 52], [0, 120], [255, 123], [123, 255]]
    >>> ct._filter_peak_list(peak_list)
    [[128, 129], [10, 52]]

    See Also
    --------
    _filter_peak_array_radius
    _filter_4D_peak_array
    _filter_peak_list_radius

    """
    peak_list_filtered = []
    for x, y in peak_list:
        if x == 0:
            pass
        elif y == 0:
            pass
        elif x == max_x_index:
            pass
        elif y == max_y_index:
            pass
        else:
            peak_list_filtered.append([x, y])
    return peak_list_filtered


def _filter_peak_array_radius(peak_array, xc, yc, r_min=None, r_max=None):
    """Remove peaks from a peak_array, based on distance from a point.

    Parameters
    ----------
    peak_array : NumPy array
        In the form [[[[y0, x0], [y1, x1]]]]
    xc, yc : scalars, NumPy array
        Centre position
    r_min, r_max : scalar
        Remove peaks which are within r_min and r_max distance from the centre.
        One of them must be specified.

    Returns
    -------
    peak_array_filtered : NumPy array
        Similar to peak_array input, but with the too-close peaks
        removed.

    See Also
    --------
    _filter_peak_list
    _filter_4D_peak_array
    _filter_peak_list_radius

    """
    if not isiterable(xc):
        xc = np.ones(peak_array.shape[:2]) * xc
    if not isiterable(yc):
        yc = np.ones(peak_array.shape[:2]) * yc
    peak_array_filtered = np.empty(shape=peak_array.shape[:2], dtype=object)
    for iy, ix in np.ndindex(peak_array.shape[:2]):
        temp_xc, temp_yc = xc[iy, ix], yc[iy, ix]
        peak_list_filtered = _filter_peak_list_radius(
            peak_array[iy, ix], xc=temp_xc, yc=temp_yc, r_min=r_min, r_max=r_max
        )
        peak_array_filtered[iy, ix] = np.array(peak_list_filtered)
    return peak_array_filtered


def _filter_peak_list_radius(peak_list, xc, yc, r_min=None, r_max=None):
    """Remove peaks based on distance to some point.

    Parameters
    ----------
    peak_list : NumPy array
        In the form [[y0, x0], [y1, x1], ...]
    xc, yc : scalars
        Centre position
    r_min, r_max : scalar
        Remove peaks which are within r_min and r_max distance from the centre.
        One of them must be specified.

    Returns
    -------
    peak_filtered_list : NumPy array
        Similar to peak_list input, but with the too-close peaks
        removed.

    Examples
    --------
    >>> import pyxem.utils.cluster_tools as ct
    >>> peak_list = np.array([[128, 32], [128, 127]])
    >>> ct._filter_peak_list_radius(peak_list, 128, 128, r_min=10)
    array([[128,  32]])

    See Also
    --------
    _filter_peak_array_radius
    _filter_peak_list
    _filter_4D_peak_array

    """
    dist = np.hypot(peak_list[:, 1] - xc, peak_list[:, 0] - yc)
    if (r_min is None) and (r_max is None):
        raise ValueError("Either r_min or r_max must be specified")
    if (r_min is not None) and (r_max is not None):
        if r_max < r_min:
            raise ValueError(
                "r_min ({0}) must be smaller than r_max ({1})".format(r_min, r_max)
            )
    filter_list = np.ones_like(dist, dtype=bool)
    if r_min is not None:
        temp_filter_list = dist > r_min
        filter_list[:] = np.logical_and(filter_list, temp_filter_list)
    if r_max is not None:
        temp_filter_list = dist < r_max
        filter_list[:] = np.logical_and(filter_list, temp_filter_list)
    peak_filtered_list = peak_list[filter_list]
    return peak_filtered_list


def _get_cluster_dict(peak_array, eps=30, min_samples=2):
    """Sort peaks into cluster using sklearn's DBSCAN.

    Each cluster is given its own label, with the unclustered
    having the label -1.

    Parameters
    ----------
    peak_array : 2D numpy array
        In the form [[x0, y0], [x1, y1], ...], i.e. shape = 2
    eps : scalar
        For the DBSCAN clustering algorithm
    min_samples : int
        Minimum number of peaks in each cluster

    Returns
    -------
    cluster_dict : dict
        The peaks are sorted into a dict with the cluster label as the key.

    Example
    -------
    >>> import numpy as np
    >>> peak_array = np.random.randint(1000, size=(100, 2))
    >>> import pyxem.utils.cluster_tools as ct
    >>> cluster_dict = ct._get_cluster_dict(peak_array)
    >>> cluster0 = cluster_dict[0]

    """
    dbscan = cluster.DBSCAN(eps=eps, min_samples=min_samples)
    dbscan.fit(peak_array)
    label_list = dbscan.labels_

    label_unique_list = sorted(list(set(label_list)))
    cluster_dict = {}
    for label_unique in label_unique_list:
        cluster_dict[label_unique] = []

    for peak, label in zip(peak_array, label_list):
        cluster_dict[label].append(peak.tolist())
    return cluster_dict


def _sort_cluster_dict(cluster_dict, centre_x=128, centre_y=128):
    """Sort clusters into centre, rest and unclustered.

    Parameters
    ----------
    cluster_dict : dict
    centre_x : scalar, optional
        Default 128
    centre_y : scalar, optional
        Default 128

    Returns
    -------
    sorted_cluster_dict : dict
        Centre cluster has key 'centre', the others 'rest',
        and lastly the unclustered 'none'.

    Examples
    --------
    >>> import numpy as np
    >>> peak_array0 = np.random.randint(6, size=(100, 2)) + 128
    >>> peak_array1 = np.random.randint(6, size=(100, 2)) + 200
    >>> peak_array = np.vstack((peak_array0, peak_array1, [[100, 0], ]))
    >>> import pyxem.utils.cluster_tools as ct
    >>> cluster_dict = ct._get_cluster_dict(peak_array)
    >>> sorted_cluster_dict = ct._sort_cluster_dict(cluster_dict)
    >>> cluster_centre = sorted_cluster_dict['centre']
    >>> cluster_rest = sorted_cluster_dict['rest']
    >>> cluster_none = sorted_cluster_dict['none']

    Different centre position

    >>> sorted_cluster_dict = ct._sort_cluster_dict(
    ...     cluster_dict, centre_x=200, centre_y=200)

    """
    label_list, closest_list = [], []
    for label, cluster_list in cluster_dict.items():
        label_list.append(label)
        cluster_array = np.array(cluster_list)
        r = np.hypot(cluster_array[:, 0] - centre_x, cluster_array[:, 1] - centre_y)
        closest_list.append(_find_nearest(r, 0))

    icentre_label = np.argmin(closest_list)
    centre_label = label_list[icentre_label]

    sorted_cluster_dict = {"none": [], "centre": [], "rest": []}
    for label, cluster_list in cluster_dict.items():
        if label == -1:
            sorted_cluster_dict["none"] = cluster_list
        elif label == centre_label:
            sorted_cluster_dict["centre"] = cluster_list
        else:
            sorted_cluster_dict["rest"].extend(cluster_list)
    return sorted_cluster_dict


def _get_peak_array_shape(peak_array):
    """Find the navigation shape of a peak array

    This is necessary due to the peak_array.shape will be different
    depending if the array is the more common object dtype, or
    something else.

    Parameters
    ----------
    peak_array : NumPy array

    Returns
    -------
    peak_array_shape : tuple

    """
    if peak_array.dtype == object:
        peak_array_shape = peak_array.shape
    else:
        peak_array_shape = peak_array.shape[:-2]
    return peak_array_shape


def _cluster_and_sort_peak_array(
    peak_array, eps=30, min_samples=2, centre_x=128, centre_y=128
):
    """Cluster and sort a 4D peak array into centre, rest and unclustered.

    Parameters
    ----------
    peak_array : NumPy array
    eps : scalar, optional
        Default 30, passed to sklearn's DBSCAN
    min_samples : scalar, optional
        Default 2, passed to sklearn's DBSCAN
    centre_x, centre_y : scalar, optional
        Default 128

    Returns
    -------
    peak_dicts : dict
        Different peaks sorted into either peak_dicts['centre'],
        peak_dicts['rest'], peak_dicts['none'] (for the unclustered points).

    Example
    -------
    >>> peak_array0 = np.random.randint(124, 132, size=(2, 4, 10, 2))
    >>> peak_array1 = np.random.randint(204, 208, size=(2, 4, 10, 2))
    >>> peak_array = np.concatenate((peak_array0, peak_array1), axis=2)
    >>> import pyxem.utils.cluster_tools as ct
    >>> peak_dicts = ct._cluster_and_sort_peak_array(peak_array)
    >>> peak_array_centre = peak_dicts['centre']
    >>> peak_array_rest = peak_dicts['rest']
    >>> peak_array_none = peak_dicts['none']

    """
    peak_array_shape = _get_peak_array_shape(peak_array)
    peak_centre_array = np.empty(shape=peak_array_shape, dtype=object)
    peak_rest_array = np.empty(shape=peak_array_shape, dtype=object)
    peak_none_array = np.empty(shape=peak_array_shape, dtype=object)
    for index in np.ndindex(peak_array_shape):
        islice = np.s_[index]
        cluster_dict = _get_cluster_dict(
            peak_array[islice], eps=eps, min_samples=min_samples
        )
        sorted_cluster_dict = _sort_cluster_dict(
            cluster_dict, centre_x=centre_x, centre_y=centre_y
        )
        peak_centre_array[islice] = sorted_cluster_dict["centre"]
        peak_rest_array[islice] = sorted_cluster_dict["rest"]
        peak_none_array[islice] = sorted_cluster_dict["none"]

    peak_dicts = {}
    peak_dicts["centre"] = peak_centre_array
    peak_dicts["rest"] = peak_rest_array
    peak_dicts["none"] = peak_none_array
    return peak_dicts


def _add_peak_dicts_to_signal(
    signal,
    peak_dicts,
    color_centre="red",
    color_rest="blue",
    color_none="cyan",
    size=20,
):
    """Visualize the results of peak_dicts through markers in a Signal.

    Parameters
    ----------
    signal : HyperSpy Signal2D, PixelatedSTEM
    peak_dicts : dicts
    color_centre, color_rest, color_none : string, optional
        Color of the markers. Default 'red', 'blue', 'cyan'.
    size : scalar, optional
        Size of the markers. Default 20

    Example
    -------
    >>> peak_dicts = {}
    >>> peak_dicts['centre'] = np.random.randint(99, size=(2, 3, 10, 2))
    >>> peak_dicts['rest'] = np.random.randint(99, size=(2, 3, 3, 2))
    >>> peak_dicts['none'] = np.random.randint(99, size=(2, 3, 2, 2))
    >>> s = pxm.signals.Diffraction2D(np.random.random((2, 3, 100, 100)))
    >>> import pyxem.utils.cluster_tools as ct
    >>> ct._add_peak_dicts_to_signal(s, peak_dicts)
    >>> s.plot()

    """
    center_points = hs.plot.markers.Points(
        peak_dicts["centre"],
        color=color_centre,
        sizes=[
            size,
        ],
    )
    rest_points = hs.plot.markers.Points(
        peak_dicts["rest"],
        color=color_rest,
        sizes=[
            size,
        ],
    )
    none_points = hs.plot.markers.Points(
        peak_dicts["none"],
        color=color_none,
        sizes=[
            size,
        ],
    )
    signal.add_marker([center_points, rest_points, none_points])


def _sorted_cluster_dict_to_marker_list(
    sorted_cluster_dict,
    signal_axes=None,
    color_centre="blue",
    color_rest="red",
    color_none="green",
    size=20,
):
    """Make a list of markers with different colors from a sorted cluster dict

    Parameters
    ----------
    sorted_cluster_dict : dict
        dict with clusters sorted into 'centre', 'rest' and 'none'
        lists.
    signal_axes : HyperSpy axes_manager, optional
    color_centre, color_rest, color_none : string, optional
        Color of the markers. Default 'blue', 'red' and 'green'.
    size : scalar, optional
        Size of the markers.

    Returns
    -------
    marker_list : list of HyperSpy markers

    Examples
    --------
    >>> from numpy.random import randint
    >>> sorted_cluster_dict = {}
    >>> center_arr = np.empty((3,4), dtype=object)
    >>> rest_arr = np.empty((3,4), dtype=object)
    >>> none_arr = np.empty((3,4), dtype=object)
    >>> for i in np.ndindex((3,4)):
    ...     center_arr[i] = randint(10, size=(10, 2))
    ...     rest_arr[i] = randint(50, 60, size=(10, 2))
    ...     none_arr[i] = randint(90, size=(2, 2))
    >>> sorted_cluster_dict['centre'] = center_arr
    >>> sorted_cluster_dict['rest'] = rest_arr
    >>> sorted_cluster_dict['none'] = none_arr
    >>> import pyxem.utils.cluster_tools as ct
    >>> marker_list = ct._sorted_cluster_dict_to_marker_list(
    ...     sorted_cluster_dict)
    >>> import pyxem.utils.marker_tools as mt
    >>> s = pxm.signals.Diffraction2D(np.random.random((3, 4, 100, 100)))
    >>> mt._add_permanent_markers_to_signal(s, marker_list)

    Different colors

    >>> marker_list = ct._sorted_cluster_dict_to_marker_list(
    ...     sorted_cluster_dict, color_centre='green', color_rest='cyan',
    ...     color_none='purple', size=15)

    """

    marker_list = []
    for label, cluster_list in sorted_cluster_dict.items():
        if label == "centre":
            color = color_centre
        elif label == "rest":
            color = color_rest
        elif label == "none":
            color = color_none
        else:
            color = "cyan"

        marker_list.append(
            hs.plot.markers.Points(cluster_list, color=color, size=size),
        )
    return marker_list



# -*- coding: utf-8 -*-
# Copyright 2016-2025 The pyXem developers
#
# This file is part of pyXem.
#
# pyXem is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyXem is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyXem.  If not, see <http://www.gnu.org/licenses/>.

import numpy as np
import pytest
from transforms3d.euler import euler2mat
import dask.array as da

import sys

from diffsims.generators.simulation_generator import SimulationGenerator
from orix.sampling import get_sample_reduced_fundamental
from orix.quaternion import Rotation, Orientation
from orix.crystal_map import CrystalMap
from orix.vector import Vector3d

from pyxem.signals import VectorMatchingResults, DiffractionVectors, OrientationMap
from pyxem.utils.indexation_utils import OrientationResult
from pyxem.data import (
    si_grains,
    si_phase,
    si_tilt,
    si_grains_simple,
    fe_multi_phase_grains,
    fe_bcc_phase,
    fe_fcc_phase,
)

from pyxem.signals.indexation_results import vectors_from_orientation_map, phase2dict
import hyperspy.api as hs


@pytest.fixture(scope="module", autouse=True)
def single_threaded_on_mac():
    """Ensure tests run single-threaded if executed on macOS.

    This is a workaround, likely related to numpy/numba wheels from pypi.
    See pyxem/pull/1138 for more details.
    """
    if sys.platform == "darwin":
        import dask

        dask.config.set(scheduler="single-threaded")


@pytest.fixture
def sp_vector_match_result():
    # We require (total_error of row_1 > correlation row_2)
    res = np.empty(2, dtype="object")
    res[0] = OrientationResult(
        0,
        euler2mat(*np.deg2rad([0, 0, 90]), "rzxz"),
        0.5,
        np.array([0.1, 0.05, 0.2]),
        0.1,
        1.0,
        0,
        0,
    )
    res[1] = OrientationResult(
        0,
        euler2mat(*np.deg2rad([0, 0, 90]), "rzxz"),
        0.6,
        np.array([0.1, 0.10, 0.2]),
        0.2,
        1.0,
        0,
        0,
    )
    return VectorMatchingResults(res)


@pytest.fixture
def dp_vector_match_result():
    res = np.empty(4, dtype="object")
    res = res.reshape(2, 2)
    res[0, 0] = OrientationResult(
        0,
        euler2mat(*np.deg2rad([90, 0, 0]), "rzxz"),
        0.6,
        np.array([0.1, 0.10, 0.2]),
        0.3,
        1.0,
        0,
        0,
    )
    res[0, 1] = OrientationResult(
        0,
        euler2mat(*np.deg2rad([0, 10, 20]), "rzxz"),
        0.5,
        np.array([0.1, 0.05, 0.2]),
        0.4,
        1.0,
        0,
        0,
    )
    res[1, 0] = OrientationResult(
        1,
        euler2mat(*np.deg2rad([0, 45, 45]), "rzxz"),
        0.8,
        np.array([0.1, 0.30, 0.2]),
        0.1,
        1.0,
        0,
        0,
    )
    res[1, 1] = OrientationResult(
        1,
        euler2mat(*np.deg2rad([0, 0, 90]), "rzxz"),
        0.7,
        np.array([0.1, 0.05, 0.1]),
        0.2,
        1.0,
        0,
        0,
    )
    return VectorMatchingResults(res)


@pytest.mark.skip(reason="This functionality is under limited support as of 0.14.0")
def test_single_vector_get_crystallographic_map(sp_vector_match_result):
    _ = sp_vector_match_result.get_crystallographic_map()


@pytest.mark.skip(reason="This functionality is under limited support as of 0.14.0")
def test_double_vector_get_crystallographic_map(dp_vector_match_result):
    _ = dp_vector_match_result.get_crystallographic_map()


@pytest.mark.parametrize(
    "overwrite, result_hkl, current_hkl, expected_hkl",
    [
        (True, [0, 0, 1], None, [0, 0, 1]),
        (False, [0, 0, 1], None, [0, 0, 1]),
    ],
)
def test_vector_get_indexed_diffraction_vectors(
    overwrite, result_hkl, current_hkl, expected_hkl
):
    match_results = VectorMatchingResults(np.array([[1], [2]]))
    match_results.hkls = result_hkl
    vectors = DiffractionVectors(np.array([[1], [2]]))
    vectors.hkls = current_hkl
    match_results.get_indexed_diffraction_vectors(vectors, overwrite)
    np.testing.assert_allclose(vectors.hkls, expected_hkl)


def test_vector_get_indexed_diffraction_vectors_warn():
    match_results = VectorMatchingResults(np.array([[1], [2]]))
    match_results.hkls = [0, 0, 1]
    vectors = DiffractionVectors(np.array([[1], [2]]))
    vectors.hkls = [0, 0, 0]
    with pytest.warns(Warning):
        match_results.get_indexed_diffraction_vectors(vectors)
    np.testing.assert_allclose(vectors.hkls, [0, 0, 0])


class TestOrientationResult:
    """Testing the OrientationMap class for valid outputs. These tests are based on the
    examples provided in the documentation.
    """

    @pytest.fixture(scope="class")
    def single_rot_orientation_result(self):
        s = si_tilt()
        s.calibration.center = None
        polar_si_tilt = s.get_azimuthal_integral2d(
            npt=100, npt_azim=360, inplace=False, mean=True
        )
        phase = si_phase()
        generator = SimulationGenerator(200)
        sim = generator.calculate_diffraction2d(
            phase,
            rotation=Rotation.from_euler(
                [0, 0, 0],
                degrees=True,
            ),
            max_excitation_error=0.1,
            reciprocal_radius=1.5,
            with_direct_beam=False,
        )
        orientation_map = polar_si_tilt.get_orientation(sim)
        return orientation_map

    @pytest.fixture(scope="class")
    def multi_rot_orientation_result(self):
        s, r = si_grains(return_rotations=True)
        s.calibration.center = None
        polar = s.get_azimuthal_integral2d(
            npt=100, npt_azim=180, inplace=False, mean=True
        )
        phase = si_phase()
        generator = SimulationGenerator(200, minimum_intensity=0.05)
        rotations = get_sample_reduced_fundamental(
            resolution=1, point_group=phase.point_group
        )
        sims = generator.calculate_diffraction2d(
            phase,
            rotation=rotations,
            max_excitation_error=0.1,
            reciprocal_radius=2,
            with_direct_beam=False,
        )
        orientations = polar.get_orientation(sims)

        return orientations, r

    @pytest.fixture(scope="class")
    def simple_multi_rot_orientation_result(self):
        s, r = si_grains_simple(return_rotations=True)
        s.calibration.center = None
        polar = s.get_azimuthal_integral2d(
            npt=100, npt_azim=180, inplace=False, mean=True
        )
        phase = si_phase()
        generator = SimulationGenerator(200, minimum_intensity=0.05)
        rotations = get_sample_reduced_fundamental(
            resolution=1, point_group=phase.point_group
        )
        sims = generator.calculate_diffraction2d(
            phase,
            rotation=rotations,
            max_excitation_error=0.1,
            reciprocal_radius=2,
            with_direct_beam=True,
        )
        polar = polar**1
        orientations = polar.get_orientation(sims)
        return orientations, r, s

    @pytest.fixture(scope="class")
    def multi_phase_orientation_result(self):
        s = fe_multi_phase_grains()
        s.calibration.center = None
        polar = s.get_azimuthal_integral2d(
            npt=100, npt_azim=180, inplace=False, mean=True
        )
        phase = fe_fcc_phase()
        phase2 = fe_bcc_phase()

        generator = SimulationGenerator(200, minimum_intensity=0.05)
        rotations = get_sample_reduced_fundamental(
            resolution=2, point_group=phase.point_group
        )
        rotations2 = get_sample_reduced_fundamental(
            resolution=1, point_group=phase2.point_group
        )
        sims = generator.calculate_diffraction2d(
            [phase, phase2],
            rotation=[rotations, rotations2],
            max_excitation_error=0.1,
            reciprocal_radius=2,
        )
        orientations = polar.get_orientation(sims, n_best=3)
        return orientations

    def test_tilt_orientation_result(self, single_rot_orientation_result):
        assert isinstance(single_rot_orientation_result, OrientationMap)
        orients = single_rot_orientation_result.to_single_phase_orientations()
        # Check that the orientations are within 1 degree of the expected value
        degrees_between = orients.angle_with(
            Orientation.from_euler([0, 0, 0]), degrees=True
        )
        assert np.all(
            degrees_between[:, :2] <= 1
        )  # off by 1 degree (due to pixelation?)
        degrees_between = orients.angle_with(
            Orientation.from_euler([10, 0, 0], degrees=True), degrees=True
        )
        assert np.all(degrees_between[:, 2:] <= 1)

    def test_grain_orientation_result(self, simple_multi_rot_orientation_result):
        orientations, rotations, s = simple_multi_rot_orientation_result
        assert isinstance(rotations, Orientation)
        assert isinstance(orientations, OrientationMap)
        orients = orientations.to_single_phase_orientations()

        v1 = (orients * Vector3d.zvector()).in_fundamental_sector(orients.symmetry)
        v2 = (rotations * Vector3d.zvector()).in_fundamental_sector(rotations.symmetry)

        # Check that the orientations are within 2 degrees of the expected value.
        # Use 2 degrees since that is the angular resolution of the polar dataset
        degrees_between = v1.angle_with(v2, degrees=True)
        min_deg = np.min(degrees_between, axis=2)
        np.testing.assert_allclose(min_deg, 0, atol=3)

    def test_to_crystal_map(self, simple_multi_rot_orientation_result):
        orientations, rotations, s = simple_multi_rot_orientation_result
        crystal_map = orientations.to_crystal_map()
        assert isinstance(crystal_map, CrystalMap)
        assert np.all(crystal_map.phase_id == 0)

    def test_to_crystal_map_multi_phase(self, multi_phase_orientation_result):
        crystal_map = multi_phase_orientation_result.to_crystal_map()
        assert isinstance(crystal_map, CrystalMap)
        assert np.all(crystal_map.phase_id < 2)

    @pytest.mark.parametrize("annotate", [True, False])
    @pytest.mark.parametrize("lazy_output", [True, False])
    @pytest.mark.parametrize("add_intensity", [True, False])
    def test_to_markers(
        self,
        simple_multi_rot_orientation_result,
        annotate,
        lazy_output,
        add_intensity,
    ):
        fast = True
        orientations, rotations, s = simple_multi_rot_orientation_result
        markers = orientations.to_markers(
            lazy_output=lazy_output,
            annotate=annotate,
            include_intensity=add_intensity,
            fast=fast,
        )
        assert isinstance(markers[0], hs.plot.markers.Markers)

    def test_to_markers_lazy(self, simple_multi_rot_orientation_result):
        orientations, rotations, s = simple_multi_rot_orientation_result
        orientations = orientations.as_lazy()
        markers = orientations.to_markers()
        assert isinstance(markers[0].kwargs["offsets"], da.Array)

    def test_to_single_phase_markers_polar(self, simple_multi_rot_orientation_result):
        orientations, rotations, s = simple_multi_rot_orientation_result
        polar = s.get_azimuthal_integral2d(
            npt=100, npt_azim=180, inplace=False, mean=True
        )
        from pyxem.common import VisibleDeprecationWarning

        with pytest.warns(VisibleDeprecationWarning):
            markers = orientations.to_single_phase_polar_markers(
                polar.axes_manager.signal_axes
            )
        assert isinstance(markers[0], hs.plot.markers.Markers)

    def test_to_markers_polar(self, simple_multi_rot_orientation_result):
        orientations, rotations, s = simple_multi_rot_orientation_result
        markers = orientations.to_polar_markers()
        assert isinstance(markers[0], hs.plot.markers.Markers)

    def test_to_ipf_markers(self, simple_multi_rot_orientation_result):
        orientations, rotations, s = simple_multi_rot_orientation_result
        markers = orientations.to_ipf_markers()
        assert isinstance(markers[0], hs.plot.markers.Markers)

    @pytest.mark.parametrize("add_markers", [True, False])
    def test_to_ipf_map(self, simple_multi_rot_orientation_result, add_markers):
        orientations, rotations, s = simple_multi_rot_orientation_result
        navigator = orientations.to_ipf_colormap(add_markers=add_markers)
        assert isinstance(navigator, hs.signals.BaseSignal)
        if add_markers:
            assert len(navigator.metadata.Markers) == 3

    def test_to_phasemap(self, multi_phase_orientation_result):
        navigator = multi_phase_orientation_result.to_phase_map()
        assert isinstance(navigator, hs.signals.BaseSignal)

    def test_multi_phase_errors(self, multi_phase_orientation_result):
        with pytest.raises(ValueError):
            multi_phase_orientation_result.to_ipf_colormap()
        with pytest.raises(ValueError):
            multi_phase_orientation_result.to_single_phase_orientations()

    def test_lazy_error(self, simple_multi_rot_orientation_result):
        orientations, rotations, s = simple_multi_rot_orientation_result
        orientations = orientations.as_lazy()
        with pytest.raises(ValueError):
            rotations = orientations.to_rotation()
        with pytest.raises(ValueError):
            rotations = orientations.to_ipf_markers()

    def test_to_crystal_map_error(self, simple_multi_rot_orientation_result):
        orientations, rotations, s = simple_multi_rot_orientation_result
        orientations = hs.stack((orientations, orientations))
        with pytest.raises(ValueError):
            rotations = orientations.to_crystal_map()

    @pytest.mark.parametrize("add_vector_markers", [False, True])
    @pytest.mark.parametrize("add_ipf_markers", [False, True])
    @pytest.mark.parametrize("add_ipf_correlation_heatmap", [False, True])
    @pytest.mark.parametrize("add_ipf_colorkey", [False, True])
    @pytest.mark.parametrize(
        "vector_kwargs", [None, {"annotate": False}, {"annotate": True}]
    )
    def test_plot_over_single_phase_signal(
        self,
        simple_multi_rot_orientation_result,
        add_vector_markers,
        add_ipf_markers,
        add_ipf_correlation_heatmap,
        add_ipf_colorkey,
        vector_kwargs,
    ):
        orientations, rotations, s = simple_multi_rot_orientation_result
        orientations.plot_over_signal(
            s,
            add_vector_markers=add_vector_markers,
            add_ipf_markers=add_ipf_markers,
            add_ipf_correlation_heatmap=add_ipf_correlation_heatmap,
            add_ipf_colorkey=add_ipf_colorkey,
            vector_kwargs=vector_kwargs,
        )

    @pytest.mark.parametrize("add_vector_markers", [False, True])
    @pytest.mark.parametrize("add_ipf_correlation_heatmap", [False, True])
    @pytest.mark.parametrize("add_ipf_colorkey", [False, True])
    @pytest.mark.parametrize(
        "vector_kwargs", [None, {"annotate": False}, {"annotate": True}]
    )
    def test_plot_over_multi_phase_signal(
        self,
        multi_phase_orientation_result,
        add_vector_markers,
        add_ipf_correlation_heatmap,
        add_ipf_colorkey,
        vector_kwargs,
    ):
        add_ipf_markers = True
        # Mock signal
        s = hs.signals.Signal2D(
            np.zeros(multi_phase_orientation_result.data.shape[:-2])[
                ..., np.newaxis, np.newaxis
            ]
        )
        multi_phase_orientation_result.plot_over_signal(
            s,
            add_vector_markers=add_vector_markers,
            add_ipf_markers=add_ipf_markers,
            add_ipf_correlation_heatmap=add_ipf_correlation_heatmap,
            add_ipf_colorkey=add_ipf_colorkey,
            vector_kwargs=vector_kwargs,
        )

    def test_to_ipf_correlation_heatmap_markers_single_phase(
        self, simple_multi_rot_orientation_result
    ):
        orientations, rotations, s = simple_multi_rot_orientation_result
        markers = orientations.to_ipf_correlation_heatmap_markers()
        assert all(isinstance(m, hs.plot.markers.Markers) for m in markers)

    def test_to_ipf_correlation_heatmap_markers_multi_phase(
        self, multi_phase_orientation_result
    ):
        markers = multi_phase_orientation_result.to_ipf_correlation_heatmap_markers()
        assert all(isinstance(m, hs.plot.markers.Markers) for m in markers)

    def test_vector_markers_correctness(self):
        """
        Check if the markers are plotted correctly by performing orientation mapping
        on a non-centrosymmetric signal, where only one quadrant is non-zero,
        using a template with only one diffraction spot.
        This spot should then be plotted in the correct quadrant.
        """
        from pyxem.signals import Diffraction2D

        # Simple signal: 1 in first quadrant, 0 elsewhere
        signal = Diffraction2D(np.array([[[[0, 1], [0, 0]]]]))
        signal.calibration(center=None)
        polar = signal.get_azimuthal_integral2d(npt=10, npt_azim=36)

        from diffpy.structure import Lattice, Atom, Structure
        from orix.crystal_map import Phase

        # Primitive cubic structure, any will do
        l = Lattice(5, 5, 5, 90, 90, 90)
        a = [Atom("Au", xyz=[0, 0, 0], lattice=l)]
        s = Structure(a, l)
        p = Phase(space_group=221, structure=s)
        gen = SimulationGenerator()
        sim = gen.calculate_diffraction2d(p, with_direct_beam=False)

        # Set intensities to 0 for all vectors except one
        _, _, v = sim.get_simulation(0)
        v.intensity[:-1] = 0
        v.intensity[-1] = 1

        # Perform matching
        res = polar.get_orientation(sim)

        # Check cartesian
        x, y = res.to_markers()[0].kwargs["offsets"][0, 0][-1]
        assert x > 0
        # When plotting, the y-axis points downwards, but pyxem defines it as upwards.
        # Therefore, the first quadrant, as pyxem defines it, has negative y when plotting.
        assert y < 0

        # Check polar
        r, t = res.to_polar_markers()[0].kwargs["offsets"][0, 0][-1]
        assert r > 0
        assert 0 < t < np.pi / 2

    def test_vectors_from_orientation_map(self, single_rot_orientation_result):
        data, intensities, phases, phase_indices, hkl = (
            single_rot_orientation_result.get_simulation_arrays()
        )
        phases_dicts = [phase2dict(p) for p in phases]
        original_data = data.copy()
        v = vectors_from_orientation_map(
            single_rot_orientation_result.data[0, 0],
            vectors=data,
            phases=phases_dicts,
            hkl=hkl,
            phase_index=phase_indices,
            intensities=intensities,
            n_best_index=0,
        )

        np.testing.assert_almost_equal(
            v.original_hkl, single_rot_orientation_result.simulation.coordinates.hkl
        )
        np.testing.assert_equal(original_data, data)



# -*- coding: utf-8 -*-
# Copyright 2016-2025 The pyXem developers
#
# This file is part of pyXem.
#
# pyXem is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyXem is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyXem.  If not, see <http://www.gnu.org/licenses/>.

import numpy as np
import pytest

from pyxem.utils.segment_utils import (
    norm_cross_corr,
    separate_watershed,
    get_gaussian2d,
)


@pytest.mark.parametrize(
    "img, template, corr_expt",
    [
        (
            np.array([[0.1, 50.0, 0.0], [0.0, 0.0, 44], [0.3, 37, 11.1]]),
            np.array([[0.2, 0.0, 1.4], [10.9, 22.0, 15.0], [0.28, 36, 0.36]]),
            0.28918685246563913,
        ),
        (
            np.array([[1, 9, 5], [42, 42, 44], [6, 37, 11]]),
            np.array([[0, 1, 0], [1, 0, 0], [40, 90, 35]]),
            0.075079495399324,
        ),
        (
            np.array([[1, 0, 1], [0, 1, 1], [1, 0, 0]]),
            np.array([[0, 1, 0], [1, 0, 0], [0, 1, 1]]),
            -1.0,
        ),
        (
            np.array([[1, 9, 5], [42, 42, 44], [6, 37, 11]]),
            np.array([[1, 9, 5], [42, 42, 44], [6, 37, 11]]),
            1.0,
        ),
    ],
)
def test_norm_cross_corr(img, template, corr_expt):
    c = norm_cross_corr(img, template)
    np.testing.assert_allclose(c, corr_expt)


@pytest.fixture
def vdf_image():
    stest = np.zeros((7, 6))
    stest[1:3, :2] = 1
    stest[2:5, 1:3] = 2
    stest[1:3, 4:] = 3
    stest[4, 4:] = 4
    stest[6:, :] = 5
    return stest


@pytest.mark.parametrize(
    "min_distance, min_size, max_size, "
    "max_number_of_grains, exclude_border, threshold, "
    "sep_shape_expt",
    [
        (1, 1, np.inf, np.inf, 0, True, (4, 6, 7)),
        (1, 1, np.inf, np.inf, 1, False, (1, 6, 7)),
        (1, 5, np.inf, np.inf, 0, True, (2, 6, 7)),
        (1, 1, np.inf, 4, 0, True, (3, 6, 7)),
        (1, 1, 3, np.inf, 0, True, (1, 6, 7)),
        (3, 1, np.inf, np.inf, 0, True, (1, 6, 7)),
    ],
)
def test_separate_watershed(
    vdf_image,
    min_distance,
    min_size,
    max_size,
    max_number_of_grains,
    exclude_border,
    threshold,
    sep_shape_expt,
):
    sep = separate_watershed(
        vdf_image,
        min_distance=min_distance,
        min_size=min_size,
        max_size=max_size,
        max_number_of_grains=max_number_of_grains,
        exclude_border=exclude_border,
        threshold=threshold,
    )
    # we don't care how many clusters we generate
    assert sep.shape[1:] == sep_shape_expt[1:]


@pytest.mark.parametrize(
    "a, xo, yo, x, y, sigma, gauss_shape_expt",
    [
        (10, 5, 5, np.arange(10), np.arange(10), 0.2, (10,)),
        (
            10,
            5,
            5,
            np.tile(np.arange(10), (5, 1)),
            np.tile(np.arange(10), (5, 1)),
            0.2,
            (5, 10),
        ),
    ],
)
def test_get_gaussian2d(a, xo, yo, x, y, sigma, gauss_shape_expt):
    gauss = get_gaussian2d(a, xo, yo, x, y, sigma)
    assert isinstance(gauss, np.ndarray)
    assert gauss.dtype == float
    np.testing.assert_equal(gauss.shape, gauss_shape_expt)



# -*- coding: utf-8 -*-
# Copyright 2016-2025 The pyXem developers
#
# This file is part of pyXem.
#
# pyXem is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyXem is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyXem.  If not, see <http://www.gnu.org/licenses/>.

import numpy as np

from hyperspy.signals import Signal2D
import hyperspy.api as hs
from pyxem.utils._signals import _transfer_signal_axes


def _get_rotation_matrix(x_new):
    """Calculate the rotation matrix mapping [1,0] to x_new.

    Parameters
    ----------
    x_new : list
        Coordinates of a point on the new 'x' axis.

    Returns
    -------
    R : 2 x 2 numpy.ndarray
        The rotation matrix.
    """
    try:
        rotation_angle = np.arctan(x_new[1] / x_new[0])
    except ZeroDivisionError:  # Taking x --> y
        rotation_angle = np.deg2rad(90)

    # angle sign agrees with https://en.wikipedia.org/wiki/Rotation_matrix
    R = np.array(
        [
            [np.cos(rotation_angle), -np.sin(rotation_angle)],
            [np.sin(rotation_angle), np.cos(rotation_angle)],
        ]
    )
    return R


class StrainMap(Signal2D):
    """Class for storing strain maps.

    If created within pyxem conventions are:
    - The 'y-axis' is 90 degrees from the 'x-axis'
    - Positive rotations are anticlockwise.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # check init dimension are correct

        if "current_basis_x" in kwargs.keys():
            self.current_basis_x = kwargs["current_basis_x"]
        else:
            self.current_basis_x = [1, 0]

        self.current_basis_y = np.matmul(
            np.asarray([[0, 1], [-1, 0]]), self.current_basis_x
        )

    def rotate_strain_basis(self, x_new):
        """Rotates a strain map to a new basis.

        Parameters
        ----------
        x_new : list
            The coordinates of a point on the new 'x' axis

        Returns
        -------
        StrainMap :
            StrainMap in the new (rotated) basis.

        Notes
        -----
        Conventions are described in the class documentation.

        We follow mathmatical formalism described in:
        "https://www.continuummechanics.org/stressxforms.html" (August 2019)
        """

        def apply_rotation(transposed_strain_map, R):
            """Rotates a strain matrix to a new basis, for which R maps x_old to x_new."""
            sigmaxx_old = transposed_strain_map[0]
            sigmayy_old = transposed_strain_map[1]
            sigmaxy_old = transposed_strain_map[2]

            z = np.asarray([[sigmaxx_old, sigmaxy_old], [sigmaxy_old, sigmayy_old]])

            new = np.matmul(R.T, np.matmul(z, R))
            return [new[0, 0], new[1, 1], new[0, 1], transposed_strain_map[3]]

        def apply_rotation_complete(self, R):
            """Mapping solution to return a (unclassed) strain map in a new basis."""
            from hyperspy.api import transpose

            transposed = transpose(self)[0]
            transposed_to_new_basis = transposed.map(apply_rotation, R=R, inplace=False)
            return transposed_to_new_basis.T

        """Core functionality."""

        if self.current_basis_x != [1, 0]:
            # this takes us back to [1,0] if our current map is in a diferent basis
            R = _get_rotation_matrix(self.current_basis_x).T
            strain_map_core = apply_rotation_complete(self, R)
        else:
            strain_map_core = self

        R = _get_rotation_matrix(x_new)
        transposed_to_new_basis = apply_rotation_complete(strain_map_core, R)
        meta_dict = self.metadata.as_dictionary()

        strainmap = StrainMap(
            transposed_to_new_basis, current_basis_x=x_new, metadata=meta_dict
        )
        return _transfer_signal_axes(strainmap, self)

    def plot(
        self,
        cmap="hot",
        axes_decor="off",
        scalebar="all",
        scalebar_color="black",
        tight_layout=True,
        per_row=2,
        label=None,
    ):
        """Plot the strain map."""
        if label is None:
            label = ["e11", "e22", "e12", "theta"]
        return hs.plot.plot_images(
            self,
            per_row=2,
            label=label,
            tight_layout=True,
            cmap=cmap,
            axes_decor=axes_decor,
            scalebar=scalebar,
            scalebar_color=scalebar_color,
        )



# -*- coding: utf-8 -*-
# Copyright 2016-2025 The pyXem developers
#
# This file is part of pyXem.
#
# pyXem is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyXem is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyXem.  If not, see <http://www.gnu.org/licenses/>.

import pytest
import numpy as np
from pyxem.generators import PDFGenerator1D
from pyxem.signals.reduced_intensity1d import ReducedIntensity1D
from pyxem.signals.pair_distribution_function1d import PairDistributionFunction1D


@pytest.fixture
def reduced_intensity1d():
    data = np.ones((1, 10)) * np.arange(4).reshape(4, 1)
    data = data.reshape(2, 2, 10)
    ri = ReducedIntensity1D(data)
    return ri


def test_pdf_gen_init(reduced_intensity1d):
    pdfgen = PDFGenerator1D(reduced_intensity1d)
    assert isinstance(pdfgen, PDFGenerator1D)


def test_get_pdf(reduced_intensity1d):
    pdfgen = PDFGenerator1D(reduced_intensity1d)
    pdf = pdfgen.get_pdf(s_min=0, s_max=9)
    assert isinstance(pdf, PairDistributionFunction1D)


def test_s_limits(reduced_intensity1d):
    pdfgen = PDFGenerator1D(reduced_intensity1d)
    pdf = pdfgen.get_pdf(s_min=0)
    pdf2 = pdfgen.get_pdf(s_min=0, s_max=10)
    assert np.array_equal(pdf.data, pdf2.data)


def test_s_limit_failure(reduced_intensity1d):
    pdfgen = PDFGenerator1D(reduced_intensity1d)

    with pytest.raises(
        ValueError, match="User specified s_max is larger than the maximum"
    ):
        pdf3 = pdfgen.get_pdf(s_min=0, s_max=15)


def test_signal_size():
    spectrum = np.array([5.0, 4.0, 3.0, 2.0, 2.0, 1.0, 1.0, 1.0, 0.0, 0.0])
    ri = ReducedIntensity1D(spectrum)
    pdfgen = PDFGenerator1D(ri)
    pdf = pdfgen.get_pdf(s_min=0, s_max=10)
    assert isinstance(pdf, PairDistributionFunction1D)

    ri = ReducedIntensity1D([spectrum])
    pdfgen = PDFGenerator1D(ri)
    pdf = pdfgen.get_pdf(s_min=0, s_max=10)
    assert isinstance(pdf, PairDistributionFunction1D)

    ri = ReducedIntensity1D([[spectrum]])
    pdfgen = PDFGenerator1D(ri)
    pdf = pdfgen.get_pdf(s_min=0, s_max=10)
    assert isinstance(pdf, PairDistributionFunction1D)

    ri = ReducedIntensity1D([[[spectrum]]])
    pdfgen = PDFGenerator1D(ri)
    pdf = pdfgen.get_pdf(s_min=0, s_max=10)
    shape = pdf.data.shape
    assert shape == (1, 1, 1, 2000)

    ri = ReducedIntensity1D([[[spectrum]]])
    pdfgen = PDFGenerator1D(ri)
    pdf = pdfgen.get_pdf(s_min=0, s_max=10, r_min=0, r_max=8)
    shape = pdf.data.shape
    assert shape == (1, 1, 1, 800)



# -*- coding: utf-8 -*-
# Copyright 2016-2025 The pyXem developers
#
# This file is part of pyXem.
#
# pyXem is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyXem is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyXem.  If not, see <http://www.gnu.org/licenses/>.

import numpy as np
from scipy.linalg import polar
import math

from hyperspy.signals import Signal2D
from hyperspy.utils import stack

from pyxem.signals import StrainMap
from hyperspy._signals.lazy import LazySignal


def _polar_decomposition(image, side):
    """Perform a polar decomposition of a second rank tensor.

    Parameters
    ----------
    image : numpy.ndarray
        Matrix on which to form polar decomposition.
    side : str
        'left' or 'right' the side on which to perform polar decomposition.

    Returns
    -------
    R, U :  numpy.ndarray
        Stretch and rotation matrices obtained by polar decomposition.

    """
    return np.array(polar(image, side=side))


def _get_rotation_angle(matrix):
    """Find the rotation angle associated with a given rotation matrix.

    Parameters
    ----------
    matrix : numpy.ndarray
        A rotation matrix.

    Returns
    -------
    angle :  numpy.ndarray
        Rotation angle associated with matrix.

    """
    return np.array(-math.asin(matrix[1, 0]))


class DisplacementGradientMap(Signal2D):
    """Signal class for Tensor Fields."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # Check that the signal dimensions are (3,3) for it to be a valid
        # TensorField

    def polar_decomposition(self):
        """Perform polar decomposition on the second rank tensors describing
        the TensorField. The polar decomposition is right handed and given by
        :math:`D = RU`

        Returns
        -------
        R : TensorField
            The orthogonal matrix describing the rotation field.
        U : TensorField
            The strain tensor field.

        """
        RU = self.map(
            _polar_decomposition, side="right", inplace=False, silence_warnings=True
        )
        return RU.isig[:, :, 0], RU.isig[:, :, 1]

    def get_strain_maps(self):
        """Obtain strain maps from the displacement gradient tensor at each
        navigation position in the small strain approximation.

        Returns
        -------
        strain_results : BaseSignal
            Signal of shape < 4 | , > , navigation order is e11,e22,e12,theta
        """
        if self._lazy:
            self.compute()
        R, U = self.polar_decomposition()

        e11 = np.reciprocal(U.isig[0, 0].T) - 1
        e12 = U.isig[0, 1].T
        # e21 = U.isig[1, 0].T # Same as e12
        e22 = np.reciprocal(U.isig[1, 1].T) - 1
        theta = R.map(_get_rotation_angle, inplace=False, silence_warnings=True)
        theta = theta.transpose(2)
        strain_results = stack([e11, e22, e12, theta])
        strain_map = StrainMap(strain_results)
        if len(strain_map.axes_manager.signal_axes) == 2:
            strain_map.axes_manager.signal_axes[0].name = (
                self.axes_manager.navigation_axes[0].name
            )
            strain_map.axes_manager.signal_axes[0].scale = (
                self.axes_manager.navigation_axes[0].scale
            )
            strain_map.axes_manager.signal_axes[0].units = (
                self.axes_manager.navigation_axes[0].units
            )
            strain_map.axes_manager.signal_axes[0].offset = (
                self.axes_manager.navigation_axes[0].offset
            )

            strain_map.axes_manager.signal_axes[1].name = (
                self.axes_manager.navigation_axes[1].name
            )
            strain_map.axes_manager.signal_axes[1].scale = (
                self.axes_manager.navigation_axes[1].scale
            )
            strain_map.axes_manager.signal_axes[1].units = (
                self.axes_manager.navigation_axes[1].units
            )
            strain_map.axes_manager.signal_axes[1].offset = (
                self.axes_manager.navigation_axes[1].offset
            )
        return strain_map


class LazyDisplacementGradientMap(LazySignal, DisplacementGradientMap):
    """Lazy signal class for Tensor Fields."""

    pass



# -*- coding: utf-8 -*-
# Copyright 2016-2025 The pyXem developers
#
# This file is part of pyXem.
#
# pyXem is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyXem is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyXem.  If not, see <http://www.gnu.org/licenses/>.
import orix.crystal_map
import pooch
from pyxem.data._registry import _file_names_hash, _urls
import hyperspy.api as hs
from pathlib import Path
from pyxem.release_info import version
from typing import Optional, Union
import os

# Create a downloader object
kipper = pooch.create(
    path=pooch.os_cache("pyxem"),
    base_url="",
    version=version.replace(".dev", "+"),
    version_dev="main",
    env="PYXEM_DATA_DIR",
    registry=_file_names_hash,
    urls=_urls,
)


def au_grating(allow_download=False, **kwargs):
    """An au_grating 4-D STEM dataset used to show calibration.

    Data can be acessed from https://zenodo.org/records/11284654


    Parameters
    ----------
    **kwargs
        Keyword arguments passed to :func:`~hyperspy.api.load`.

    Examples
    --------
    >>> import pyxem as pxm
    >>> s = pxm.data.au_grating()
    >>> print(s)
    <ElectronDiffraction2D, title: , dimensions: (|254, 254)>
    >>> s.plot()
    """
    grating = Dataset("au_xgrating_100kX.hspy")
    file_path = grating.fetch_file_path(allow_download=allow_download)
    return hs.load(file_path, **kwargs)


def pdnip_glass(allow_download=False, **kwargs):  # pragma: no cover
    """A small PdNiP glass 4-D STEM dataset.

    Data can be acessed from https://zenodo.org/records/11284654

    Data liscenced under CC BY 4.0

    Parameters
    ----------
    **kwargs
        Keyword arguments passed to :func:`~hyperspy.api.load`.

    Examples
    --------
    >>> import pyxem as pxm
    >>> s = pxm.data.pdnip_glass()
    >>> print(s)
    <ElectronDiffraction2D, title: , dimensions: (128, 128|128, 128)>
    >>> s.plot()
    """
    import zarr

    pdnip = Dataset("PdNiP.zspy")
    file_path = pdnip.fetch_file_path(allow_download=allow_download)
    store = zarr.ZipStore(file_path)  # for loading from zip
    return hs.load(store, **kwargs)


def zrnb_precipitate(allow_download=False, **kwargs):  # pragma: no cover
    """A small 4-D STEM dataset of a ZrNb precipitate for strain mapping.

    Data liscenced under CC BY 4.0

    Data can be acessed from https://zenodo.org/records/11284654

    Parameters
    ----------
    **kwargs
        Keyword arguments passed to :func:`~hyperspy.api.load`.

    Examples
    --------
    >>> import pyxem as pxm
    >>> s = pxm.data.zrnb_percipitate()
    >>> print(s)
    <ElectronDiffraction2D, title: , dimensions: (128, 128|128, 128)>
    >>> s.plot()
    """
    import zarr

    zrnb = Dataset("ZrNbPercipitate.zspy")
    file_path = zrnb.fetch_file_path(allow_download=allow_download)
    store = zarr.ZipStore(file_path)  # for loading from zip
    return hs.load(store, **kwargs)


def twinned_nanowire(allow_download=False, **kwargs):  # pragma: no cover
    """A small 4-D STEM dataset of a twinned nanowire for orientation mapping.

    Data can be acessed from https://zenodo.org/records/11284654

    Parameters
    ----------
    **kwargs
        Keyword arguments passed to :func:`~hyperspy.api.load`.

    Examples
    --------
    >>> import pyxem as pxm
    >>> s = pxm.data.twinned_nanowire()
    >>> print(s)
    <Diffraction2D, title: , dimensions: (30, 100|144, 144)>
    >>> s.plot()
    """
    nanowire = Dataset("twinned_nanowire.hdf5")
    file_path = nanowire.fetch_file_path(allow_download=allow_download)
    return hs.load(file_path, **kwargs, reader="hspy")


def sample_with_g(allow_download=False, **kwargs):  # pragma: no cover
    """A small 4-D STEM dataset for orientation mapping with mulitple overlapping grains.

    Data liscenced under CC BY 4.0

    Data can be acessed from https://zenodo.org/records/11284654

    Parameters
    ----------
    **kwargs
        Keyword arguments passed to :func:`~hyperspy.api.load`.

    Examples
    --------
    >>> import pyxem as pxm
    >>> s = pxm.data.sample_with_g()
    >>> print(s)
    <ElectronDiffraction2D, title: , dimensions: (25, 26|256, 256)>
    >>> s.plot()
    """
    import zarr

    sample = Dataset("sample_with_g.zspy")
    file_path = sample.fetch_file_path(allow_download=allow_download)
    store = zarr.ZipStore(file_path)  # for loading from zip
    return hs.load(store, **kwargs)


def cuag_orientations(allow_download=False, **kwargs):  # pragma: no cover
    """A small 4-D STEM dataset of CuAg with multiple orientations of the Cu FCC phase
     for orientation mapping.

    Data liscenced under CC BY 4.0

    Data can be acessed from https://zenodo.org/records/11284654

    Parameters
    ----------
    allow_download: bool
        If True, the file will be downloaded from the repository to the local cache.
    **kwargs
        Keyword arguments passed to :func:`~hyperspy.api.load`.
    Examples
    --------
    >>> import pyxem as pxm
    >>> s = pxm.data.cuag_orientations()
    >>> print(s)
    """
    import zarr

    sample = Dataset("cuzipProcessed.zspy")
    file_path = sample.fetch_file_path(allow_download=allow_download)
    store = zarr.ZipStore(file_path)  # for loading from zip
    return hs.load(store, **kwargs)


def organic_semiconductor(allow_download=False, **kwargs):  # pragma: no cover
    """A small 4-D STEM dataset of an organic semiconductor for orientation mapping.

    Data liscenced under CC BY 4.0

    Data can be acessed from https://zenodo.org/records/11284654

    Parameters
    ----------
    allow_download: bool
        If True, the file will be downloaded from the repository to the local cache.
    **kwargs
        Keyword arguments passed to :func:`~hyperspy.api.load`.
    Examples
    --------
    >>> import pyxem as pxm
    >>> s = pxm.data.organic_semiconductor()
    >>> print(s)
    """
    import zarr

    sample = Dataset("data_processed.zspy")
    file_path = sample.fetch_file_path(allow_download=allow_download)
    store = zarr.ZipStore(file_path)  # for loading from zip
    return hs.load(store, **kwargs)


def pdcusi_insitu(allow_download=False, **kwargs):  # pragma: no cover
    """A decently sized 5D STEM dataset of a PdCuSi alloy held at 390C. This
    dataset is sliced from a larger dataset and is used to demonstrate how to operate
    on a in situ 4D STEM dataset. Note that this dataset is ~6GB compressed and about
    30GB uncompressed so it might take a while to download.

    Data liscenced under CC BY 4.0

    Data can be acessed from https://zenodo.org/records/11284654

    Parameters
    ----------
    allow_download: bool
        If True, the file will be downloaded from the repository to the local cache.
    **kwargs
        Keyword arguments passed to :func:`~hyperspy.api.load`.

    Examples
    --------
    >>> import pyxem as pxm
    >>> s = pxm.data.pdcusi_insitu()
    >>> print(s)

    """
    import zarr

    sample = Dataset("PdCuSiCrystalization-zip.zspy")
    file_path = sample.fetch_file_path(allow_download=allow_download)
    store = zarr.ZipStore(file_path)  # for loading from zip
    return hs.load(store, **kwargs)


def feal_stripes(allow_download=False, **kwargs):  # pragma: no cover
    """A small 4-D STEM dataset for doing DPC on a set of magnetic FeAl stripes

    Data can be acessed from https://zenodo.org/records/11284654

    Parameters
    ----------
    allow_download: bool
        If True, the file will be downloaded from the repository to the local cache.
    **kwargs
        Keyword arguments passed to :func:`~hyperspy.api.load`.

    Examples
    --------
    >>> import pyxem as pxm
    >>> s = pxm.data.feal_stripes()
    >>> print(s)
    """
    import zarr

    sample = Dataset("FeAl_stripes.zspy")
    file_path = sample.fetch_file_path(allow_download=allow_download)
    store = zarr.ZipStore(file_path)  # for loading from zip
    return hs.load(store, **kwargs)


def sped_ag(allow_download=False, **kwargs):  # pragma: no cover
    """A small 4-D STEM dataset of a Ag sample which is good for demonstrating the
    ML capabilities of pyxem.

    Data can be acessed from https://zenodo.org/records/11284654

    Parameters
    ----------
    allow_download: bool
        If True, the file will be downloaded from the repository to the local cache.
    **kwargs
        Keyword arguments passed to :func:`~hyperspy.api.load`.

    Examples
    --------
    >>> import pyxem as pxm
    >>> s = pxm.data.sped_ag()
    >>> print(s)

    """
    import zarr

    sample = Dataset("SPED-Ag.zspy")
    file_path = sample.fetch_file_path(allow_download=allow_download)
    store = zarr.ZipStore(file_path)  # for loading from zip
    return hs.load(store, **kwargs)


def mgo_nanocrystals(allow_download=False, **kwargs):  # pragma: no cover
    """A small 4-D STEM dataset of overlapping MgO nanocrystals

    Data can be acessed from https://zenodo.org/records/11284654

    Parameters
    ----------
    **kwargs
        Keyword arguments passed to :func:`~hyperspy.api.load`.

    Examples
    --------
    >>> import pyxem as pxm
    >>> s = pxm.data.mgo_nanocrystals()
    >>> print(s)
    <ElectronDiffraction2D, title: MgO Nano-Crystals, dimensions: (105, 105|144, 144)>
    >>> s.plot()
    """
    import zarr

    sample = Dataset("mgo_nano.zspy")
    file_path = sample.fetch_file_path(allow_download=allow_download)
    store = zarr.ZipStore(file_path)  # for loading from zip
    return hs.load(store, **kwargs)


def au_grating_20cm(allow_download=False, **kwargs):  # pragma: no cover
    """A PACBED image of Gold grating with 20cm camera length
    for demonstrating the calibration of the camera length.

    Data can be acessed from https://zenodo.org/records/14113591

    Parameters
    ----------
    allow_download: bool
        If True, the file will be downloaded from the repository to the local cache.
    **kwargs
        Keyword arguments passed to :func:`~hyperspy.api.load`.

    Examples
    --------
    >>> import pyxem as pxm
    >>> s = pxm.data.au_grating_20cm()
    >>> print(s)
    <Signal2D, title: , dimensions: (|256, 256)>
    """
    grating = Dataset("au_xgrating_20cm.tif")
    file_path = grating.fetch_file_path(allow_download=allow_download)
    return hs.load(file_path, **kwargs)


def au_phase(allow_download=False, **kwargs):  # pragma: no cover
    """A gold phase object for use in orix.

    Data can be acessed from https://zenodo.org/records/14113591
    Parameters
    ----------
    allow_download: bool
        If True, the file will be downloaded from the repository to the local cache.
    **kwargs
        Keyword arguments passed to :func:`~orix.crystal_map.Phase.from_cif`.
    Examples
    --------
    >>> import pyxem as pxm
    >>> s = pxm.data.au_phase()

    """

    sample = Dataset("au.cif")
    file_path = sample.fetch_file_path(allow_download=allow_download)
    return orix.crystal_map.Phase.from_cif(file_path, **kwargs)


def small_ptychography(allow_download=False, **kwargs):  # pragma: no cover
    """A small 4-D STEM dataset of a bilayer WS2. Each Diffraction pattern is only
    8x8 pixels so the dataset is quite small although for simple non iterative
    ptychography 8x8 pixels should be sufficient.
    """
    ptychography = Dataset("smallPtychography.hspy")
    file_path = ptychography.fetch_file_path(allow_download=allow_download)
    return hs.load(file_path, **kwargs)


def zrcual_1(allow_download=False, **kwargs):  # pragma: no cover
    """A small 4-D STEM dataset of a ZrCuAl metallic glass for FEM analysis."""
    import zarr

    sample = Dataset("pos1-1.zspy")
    file_path = sample.fetch_file_path(allow_download=allow_download)
    store = zarr.ZipStore(file_path)  # for loading from zip
    return hs.load(store, **kwargs)


def zrcual_2(allow_download=False, **kwargs):  # pragma: no cover
    """A small 4-D STEM dataset of a ZrCuAl metallic glass for FEM analysis."""
    import zarr

    sample = Dataset("pos5-1.zspy")
    file_path = sample.fetch_file_path(allow_download=allow_download)
    store = zarr.ZipStore(file_path)  # for loading from zip
    return hs.load(store, **kwargs)


def zrcual_3(allow_download=False, **kwargs):  # pragma: no cover
    """A small 4-D STEM dataset of a ZrCuAl metallic glass for FEM analysis."""
    import zarr

    sample = Dataset("pos6-1.zspy")
    file_path = sample.fetch_file_path(allow_download=allow_download)
    store = zarr.ZipStore(file_path)  # for loading from zip
    return hs.load(store, **kwargs)


class Dataset:
    file_relpath: Path
    file_package_path: Path
    file_cache_path: Path
    expected_md5_hash: str = ""

    def __init__(
        self,
        file_relpath: Union[Path, str],
    ) -> None:
        if isinstance(file_relpath, str):
            file_relpath = Path(file_relpath)
        self.file_package_path = Path(os.path.dirname(__file__)) / file_relpath

        file_relpath = file_relpath
        self.file_relpath = file_relpath
        self.file_cache_path = Path(kipper.path) / self.file_relpath

        self.expected_md5_hash = _file_names_hash[self.file_relpath_str]

    @property
    def file_relpath_str(self) -> str:
        return self.file_relpath.as_posix()

    @property
    def is_in_package(self) -> bool:
        return self.file_package_path.exists()

    @property
    def is_in_cache(self) -> bool:
        return self.file_cache_path.exists()

    @property
    def file_path(self) -> Path:
        return self.file_cache_path

    @property
    def file_path_str(self) -> str:
        return self.file_path.as_posix()

    @property
    def md5_hash(self) -> Union[str, None]:
        if self.file_path.exists():
            return pooch.file_hash(self.file_path_str, alg="md5")
        else:  # pragma: no cover
            return None

    @property
    def has_correct_hash(self) -> bool:
        return self.md5_hash == self.expected_md5_hash.split(":")[1]

    @property
    def url(self) -> Union[str, None]:
        if self.file_relpath_str in _urls:
            return _urls[self.file_relpath_str]
        else:
            return None

    def fetch_file_path(
        self, allow_download: bool = False, show_progressbar: Optional[bool] = None
    ) -> str:
        if show_progressbar is None:
            show_progressbar = hs.preferences.General.show_progressbar
        downloader = pooch.HTTPDownloader(progressbar=show_progressbar)

        if self.is_in_cache:
            if self.has_correct_hash:
                file_path = self.file_relpath_str
            elif allow_download:  # pragma: no cover
                file_path = self.file_relpath_str
            else:  # pragma: no cover
                raise ValueError(
                    f"File {self.file_path_str} must be re-downloaded from the "
                    f"repository file {self.url} to your local cache {kipper.path}. "
                    "Pass `allow_download=True` to allow this re-download."
                )
        else:  # pragma: no cover
            if allow_download:
                file_path = self.file_relpath_str
            else:
                raise ValueError(
                    f"File {self.file_relpath_str} must be downloaded from the "
                    f"repository file {self.url} to your local cache {kipper.path}. "
                    "Pass `allow_download=True` to allow this download."
                )

        return kipper.fetch(file_path, downloader=downloader)



# -*- coding: utf-8 -*-
# Copyright 2016-2025 The pyXem developers
#
# This file is part of pyXem.
#
# pyXem is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyXem is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyXem.  If not, see <http://www.gnu.org/licenses/>.


from hyperspy.signals import Signal1D
from hyperspy._signals.lazy import LazySignal

from pyxem.signals.common_diffraction import CommonDiffraction


class Diffraction1D(CommonDiffraction, Signal1D):
    """Signal class for Electron Diffraction radial profiles.

    Parameters
    ----------
    *args
        See :class:`hyperspy.api.signals.Signal1D`.
    **kwargs
        See :class:`hyperspy.api.signals.Signal1D`
    """

    _signal_type = "diffraction"

    pass


class LazyDiffraction1D(LazySignal, Diffraction1D):
    pass



# -*- coding: utf-8 -*-
# Copyright 2016-2025 The pyXem developers
#
# This file is part of pyXem.
#
# pyXem is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyXem is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyXem.  If not, see <http://www.gnu.org/licenses/>.
import hyperspy.models.model1d
import hyperspy.api as hs
import pytest

import numpy as np
import dask.array as da
from scipy.signal.windows import gaussian
from diffsims.simulations import Simulation1D
from orix.crystal_map import Phase
from diffpy.structure import Atom, Lattice, Structure

from pyxem.signals import ElectronDiffraction1D, LazyElectronDiffraction1D


class TestSimpleHyperspy:
    # Tests functions that assign to hyperspy metadata

    def test_set_experimental_parameters(self, electron_diffraction1d):
        electron_diffraction1d.set_experimental_parameters(
            accelerating_voltage=3,
            camera_length=3,
            scan_rotation=1,
            convergence_angle=1,
            rocking_angle=1,
            rocking_frequency=1,
            exposure_time=1,
        )
        assert isinstance(electron_diffraction1d, ElectronDiffraction1D)

    def test_set_scan_calibration(self, electron_diffraction1d):
        electron_diffraction1d.set_scan_calibration(19)
        assert isinstance(electron_diffraction1d, ElectronDiffraction1D)

    @pytest.mark.parametrize(
        "calibration",
        [
            1,
            0.017,
            0.5,
        ],
    )
    def test_set_diffraction_calibration(self, electron_diffraction1d, calibration):
        electron_diffraction1d.set_diffraction_calibration(calibration)
        dx = electron_diffraction1d.axes_manager.signal_axes[0]
        assert dx.scale == calibration


class TestComputeAndAsLazyElectron1D:
    def test_2d_data_compute(self):
        dask_array = da.random.random((100, 150), chunks=(50, 50))
        s = LazyElectronDiffraction1D(dask_array)
        scale0, scale1, metadata_string = 0.5, 1.5, "test"
        s.axes_manager[0].scale = scale0
        s.axes_manager[1].scale = scale1
        s.metadata.Test = metadata_string
        s.compute()
        assert isinstance(s, ElectronDiffraction1D)
        assert not hasattr(s.data, "compute")
        assert s.axes_manager[0].scale == scale0
        assert s.axes_manager[1].scale == scale1
        assert s.metadata.Test == metadata_string
        assert dask_array.shape == s.data.shape

    def test_3d_data_compute(self):
        dask_array = da.random.random((4, 10, 15), chunks=(1, 10, 15))
        s = LazyElectronDiffraction1D(dask_array)
        s.compute()
        assert isinstance(s, ElectronDiffraction1D)
        assert dask_array.shape == s.data.shape

    def test_2d_data_as_lazy(self):
        data = np.random.random((100, 150))
        s = ElectronDiffraction1D(data)
        scale0, scale1, metadata_string = 0.5, 1.5, "test"
        s.axes_manager[0].scale = scale0
        s.axes_manager[1].scale = scale1
        s.metadata.Test = metadata_string
        s_lazy = s.as_lazy()
        assert isinstance(s_lazy, LazyElectronDiffraction1D)
        assert hasattr(s_lazy.data, "compute")
        assert s_lazy.axes_manager[0].scale == scale0
        assert s_lazy.axes_manager[1].scale == scale1
        assert s_lazy.metadata.Test == metadata_string
        assert data.shape == s_lazy.data.shape

    def test_3d_data_as_lazy(self):
        data = np.random.random((4, 10, 15))
        s = ElectronDiffraction1D(data)
        s_lazy = s.as_lazy()
        assert isinstance(s_lazy, LazyElectronDiffraction1D)
        assert data.shape == s_lazy.data.shape


class TestDecomposition:
    def test_decomposition_class_assignment(self, electron_diffraction1d):
        electron_diffraction1d.decomposition()
        assert isinstance(electron_diffraction1d, ElectronDiffraction1D)


class TestModelFitting:
    @pytest.fixture()
    def simple_1d_plus_model(self):
        peaks = [20, 40, 60, 80]
        x = np.zeros(100)
        for i, p in enumerate(peaks):
            g = gaussian(100, 3)
            g = np.roll(g, p, axis=0)
            if i == 0:
                g = g * 2
            x += g
        s = ElectronDiffraction1D(x)
        p = Phase(
            name="al",
            space_group=225,
            structure=Structure(
                atoms=[Atom("al", [0, 0, 0])],
                lattice=Lattice(0.405, 0.405, 0.405, 90, 90, 90),
            ),
        )
        shifted_peaks = (np.array([20, 40, 60, 80]) + 50) % 101
        shifted_peaks = np.sort(shifted_peaks)
        sim = Simulation1D(
            phase=p,
            reciprocal_spacing=shifted_peaks,
            intensities=(1, 1, 2, 1),
            hkl=[[1, 1, 1], [2, 0, 0], [3, 1, 1], [4, 0, 0]],
            reciprocal_radius=100,
            wavelength=1e-9,
        )
        model = s.model_simulation1d(sim)
        return s, model, shifted_peaks, sim

    def test_create_model(self, simple_1d_plus_model):
        s, model, peaks, _ = simple_1d_plus_model
        assert isinstance(model, hyperspy.models.model1d.Model1D)
        assert len(model) == 5
        centers = np.sort(
            [
                m.centre.value
                for m in model
                if isinstance(m, hs.model.components1D.Gaussian)
            ]
        )
        np.testing.assert_array_almost_equal(centers, peaks, decimal=1)

    def test_model2theta_scale(self, simple_1d_plus_model):
        s, model, peaks, sim = simple_1d_plus_model
        scale = s.model2theta_scale(sim, model, 300)
        np.testing.assert_almost_equal(scale, 0.012768541396519572, decimal=6)

    def test_camera_length(self, simple_1d_plus_model):
        s, model, peaks, sim = simple_1d_plus_model
        scale = s.model2camera_length(sim, model, 300, 0.051077)
        np.testing.assert_almost_equal(scale, 4, decimal=4)



# -*- coding: utf-8 -*-
# Copyright 2016-2025 The pyXem developers
#
# This file is part of pyXem.
#
# pyXem is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyXem is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyXem.  If not, see <http://www.gnu.org/licenses/>.

"""Reduced intensity generator and associated tools."""
import numpy as np

from pyxem.components import ScatteringFitComponentXTables, ScatteringFitComponentLobato
from pyxem.utils.ri_utils import subtract_pattern, mask_from_pattern

scattering_factor_dictionary = {
    "lobato": ScatteringFitComponentLobato,
    "xtables": ScatteringFitComponentXTables,
}


class ReducedIntensityGenerator1D:
    """Generates a reduced intensity 1D profile for a specified diffraction radial profile.

    Parameters
    ----------
    signal : ElectronDiffraction1D
        An electron diffraction radial profile.
    """

    def __init__(self, signal, *args, **kwargs):
        self.signal = signal
        self.cutoff = [0, signal.axes_manager.signal_axes[0].size - 1]
        self.nav_size = [
            signal.axes_manager.navigation_axes[0].size,
            signal.axes_manager.navigation_axes[1].size,
        ]
        self.sig_size = [signal.axes_manager.signal_axes[0].size]
        self.background_fit = None  # added in one of the fits below.
        self.normalisation = None

    def set_diffraction_calibration(self, calibration):
        """Defines calibration for the signal axis variable s in terms of
        A^-1 per pixel. Note that s is defined here as
        s = 2 sin(theta)/lambda = 1/d, where theta is the scattering angle,
        lambda the wavelength, and d the reciprocal spacing.

        Parameters
        ----------
        calibration: float
                    Scattering vector calibration in terms of A^-1 per pixel.
        """
        self.signal.axes_manager.signal_axes[0].scale = calibration

    def set_diffraction_offset(self, offset):
        """Defines the offset for the signal axis variable s in terms of
        A^-1 per pixel. Note that s is defined here as
        s = 2 sin(theta)/lambda = 1/d, where theta is the scattering angle,
        lambda the wavelength, and d the reciprocal spacing.

        Parameters
        ----------
        offset: float
                    Scattering vector offset in terms of A^-1 per pixel.
        """
        self.signal.axes_manager.signal_axes[0].offset = offset

    def set_s_cutoff(self, s_min, s_max):
        """Scattering vector cutoff for the purposes of fitting an atomic scattering
        factor to the 1D profile. Specified in terms of s (in inverse angstroms).
        s is defined as s = 2 sin(theta)/lambda = 1/d, where theta is the
        scattering angle, lambda the wavelength, and d the reciprocal spacing.

        Parameters
        ----------
        s_min: float
                    Minimum scattering vector amplitude for cutoff.
        s_max: float
                    Maximum scattering vector amplitude for cutoff.
        """
        self.cutoff = [s_min, s_max]
        return

    def fit_atomic_scattering(
        self,
        elements,
        fracs,
        N=1.0,
        C=0.0,
        scattering_factor="lobato",
        plot_fit=True,
        *args,
        **kwargs,
    ):
        """Fits a diffraction intensity profile to the background.

        Uses FIT = N * sum(ci * (fi^2) + C)

        The cutoff for the scattering factor fit to s is defined via the function
        set_s_cutoff above.

        Parameters
        ----------
        elements: list of str
                    A list of elements present (by symbol). No order is necessary.
                    Example: ['Ca', 'C', 'O'] (for CaCO3)
        fracs: list of float
                    A list of fraction of the respective elements. Should sum to 1.
                    Example: [0.2, 0.2, 0.6] (for CaCO3)
        N : float
                    The "slope" of the fit. Initial value is used to start the
                    fit.
        C : float
                    An additive constant to the fit. Initial value is used to
                    start the fit.
        scattering_factor : str
                    Type of scattering parameters fitted. Default is lobato.
                    See scattering_fit_component for more details.
        plot_fit: bool
                    A bool to decide if the fit from scattering is plotted
                    after fitting.
        *args:
            Arguments to be passed to hs.multifit().
        **kwargs:
            Keyword arguments to be passed to hs.multifit().
        """

        fit_model = self.signal.create_model()
        background = scattering_factor_dictionary[scattering_factor](
            elements, fracs, N, C
        )

        fit_model.append(background)
        fit_model.set_signal_range(self.cutoff)
        fit_model.multifit(*args, **kwargs)
        fit_model.reset_signal_range()
        if plot_fit is True:
            fit_model.plot()
        fit = fit_model.as_signal()

        N_values = background.N.as_signal()
        square_sum = background.square_sum

        x_size = N_values.data.shape[0]
        y_size = N_values.data.shape[1]
        normalisation = N_values.data.reshape(x_size, y_size, 1) * square_sum

        self.normalisation = normalisation
        self.background_fit = fit
        return

    def subtract_bkgd_pattern(self, bkgd_pattern, inplace=True, *args, **kwargs):
        """Subtracts a background pattern from the signal.

        This method will edit self.signal.

        Parameters
        ----------
        bkgd_pattern : np.array
            A numpy array of a single line profile of the same resolution
            (same number of pixels) as the radial profile.
        inplace : bool
            If True (default), this signal is overwritten. Otherwise, returns a
            new signal.
        *args:
            Arguments to be passed to map().
        **kwargs:
            Keyword arguments to be passed to map().
        """
        return self.signal.map(
            subtract_pattern, pattern=bkgd_pattern, inplace=inplace, *args, **kwargs
        )

    def mask_from_bkgd_pattern(
        self, mask_pattern, mask_threshold=1, inplace=True, *args, **kwargs
    ):
        """Uses a background pattern with a threshold, and sets that part of
        the signal to zero, effectively adding a mask. This can be used to mask
        the central beam.

        Parameters
        ----------
        mask_pattern : np.array
            A numpy array line profile of the same resolution
            as the radial profile.
        mask_threshold : int or float
            An integer or float threshold. Any pixel in the
            mask_pattern with lower intensity is kept, any with
            higher or equal is set to zero.
        inplace : bool
            If True (default), this signal is overwritten. Otherwise, returns a
            new signal.
        *args:
            Arguments to be passed to map().
        **kwargs:
            Keyword arguments to be passed to map().
        """

        mask_array = mask_pattern < mask_threshold

        return self.signal.map(
            mask_from_pattern,
            pattern=mask_array.astype(float),
            inplace=inplace,
            *args,
            **kwargs,
        )

    def mask_reduced_intensity(self, mask_pattern, inplace=True, *args, **kwargs):
        """Masks the reduced intensity signal by multiplying it with a pattern
        consisting of only zeroes and ones. This can be used to mask
        the central beam.

        Parameters
        ----------
        mask_pattern : np.array of 0s and 1s
            A numpy array line profile of the same resolution as the radial profile.
            Must consist only of zeroes and ones. Ones are kept while zeroes are
            set to zero.
        mask_threshold : int or float
            An integer or float threshold. Any pixel in the
            mask_pattern with lower intensity is kept, any with
            higher or equal is set to zero.
        inplace : bool
            If True (default), this signal is overwritten. Otherwise, returns a
            new signal.
        *args:
            Arguments to be passed to map().
        **kwargs:
            Keyword arguments to be passed to map().

        """

        mask_array = mask_pattern.astype(np.uint8)
        if np.max(mask_array) != 1 or np.min(mask_array) != 0:
            raise ValueError("Masking array does not consist of zeroes and ones.")

        return self.signal.map(
            mask_from_pattern, pattern=mask_array, inplace=inplace, *args, **kwargs
        )

    def get_reduced_intensity(self):
        """Obtains a reduced intensity profile from the radial profile.

        Parameters
        ----------
        s_cutoff : list of float
                    A list of the form [s_min, s_max] to change the s_cutoff
                    from the fit.

        Returns
        -------
        ri : ReducedIntensity1D

        """

        s = np.arange(self.signal.axes_manager.signal_axes[0].size, dtype="float64")
        s *= self.signal.axes_manager.signal_axes[0].scale

        ri = self.signal._deepcopy_with_new_data(
            (
                2
                * np.pi
                * s
                * np.divide(
                    (self.signal.data - self.background_fit), self.normalisation
                )
            )
        )
        ri.set_signal_type("reduced_intensity")
        title = self.signal.metadata.General.title
        ri.metadata.General.title = f"Reduce intensity of {title}"

        return ri



# -*- coding: utf-8 -*-
# Copyright 2016-2025 The pyXem developers
#
# This file is part of pyXem.
#
# pyXem is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyXem is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyXem.  If not, see <http://www.gnu.org/licenses/>.

import pytest
import numpy as np
import dask.array as da

from pyxem.signals import Correlation2D, LazyCorrelation2D, Power2D


class TestComputeAndAsLazy2D:
    def test_2d_data_compute(self):
        dask_array = da.random.random((100, 150), chunks=(50, 50))
        s = LazyCorrelation2D(dask_array)
        scale0, scale1, metadata_string = 0.5, 1.5, "test"
        s.axes_manager[0].scale = scale0
        s.axes_manager[1].scale = scale1
        s.metadata.Test = metadata_string
        s.compute()
        assert s.__class__ == Correlation2D
        assert not hasattr(s.data, "compute")
        assert s.axes_manager[0].scale == scale0
        assert s.axes_manager[1].scale == scale1
        assert s.metadata.Test == metadata_string
        assert dask_array.shape == s.data.shape

    def test_4d_data_compute(self):
        dask_array = da.random.random((4, 4, 10, 15), chunks=(1, 1, 10, 15))
        s = LazyCorrelation2D(dask_array)
        s.compute()
        assert s.__class__ == Correlation2D
        assert dask_array.shape == s.data.shape

    def test_2d_data_as_lazy(self):
        data = np.random.random((100, 150))
        s = Correlation2D(data)
        scale0, scale1, metadata_string = 0.5, 1.5, "test"
        s.axes_manager[0].scale = scale0
        s.axes_manager[1].scale = scale1
        s.metadata.Test = metadata_string
        s_lazy = s.as_lazy()
        assert s_lazy.__class__ == LazyCorrelation2D
        assert hasattr(s_lazy.data, "compute")
        assert s_lazy.axes_manager[0].scale == scale0
        assert s_lazy.axes_manager[1].scale == scale1
        assert s_lazy.metadata.Test == metadata_string
        assert data.shape == s_lazy.data.shape

    def test_4d_data_as_lazy(self):
        data = np.random.random((4, 10, 15))
        s = Correlation2D(data)
        s_lazy = s.as_lazy()
        assert s_lazy.__class__ == LazyCorrelation2D
        assert data.shape == s_lazy.data.shape


class TestGetPower:
    @pytest.fixture
    def flat_pattern(self):
        pd = Correlation2D(data=np.ones(shape=(2, 2, 25, 25)))
        pd.axes_manager.signal_axes[0].scale = 0.5
        pd.axes_manager.signal_axes[0].name = "theta"
        pd.axes_manager.signal_axes[1].scale = 2
        pd.axes_manager.signal_axes[1].name = "k"
        return pd

    def test_power_signal(self, flat_pattern):
        power = flat_pattern.get_angular_power()
        assert isinstance(power, Power2D)

    def test_power_signal_inplace(self, flat_pattern):
        power = flat_pattern.get_angular_power(inplace=True)
        assert isinstance(flat_pattern, Power2D)
        assert power is None

    def test_summed_power_signal(self, flat_pattern):
        power = flat_pattern.get_summed_angular_power(inplace=False)
        assert isinstance(power, Power2D)

    def test_summed_power_signal_inplace(self, flat_pattern):
        power = flat_pattern.get_summed_angular_power(inplace=True)
        assert isinstance(flat_pattern, Power2D)
        assert power is None

    @pytest.mark.parametrize("method", ["max", "first", "average"])
    def test_symmetry_stem_max_first(self, flat_pattern, method):
        sym_coeff = flat_pattern.get_symmetry_coefficient(method=method)
        np.testing.assert_array_almost_equal(sym_coeff.data, sym_coeff.data[0, 0, 0, 0])

    def test_symmetry_stem_max(self, flat_pattern):
        sym_coeff = flat_pattern.get_symmetry_coefficient(method="max")
        np.testing.assert_array_almost_equal(sym_coeff.data, sym_coeff.data[0, 0, 0, 0])

    def test_symmetry_stem_average(self, flat_pattern):
        sym_coeff = flat_pattern.get_symmetry_coefficient(
            method="average", angular_range=0.01, normalize=True
        )
        np.testing.assert_array_almost_equal(sym_coeff.data, sym_coeff.data[0, 0, 0, 0])

    def test_symmetry_stem_lazy(self, flat_pattern):
        flat_pattern = flat_pattern.as_lazy()
        sym_coeff = flat_pattern.get_symmetry_coefficient(
            method="average", angular_range=0.01, normalize=True
        )
        assert isinstance(sym_coeff.data, da.Array)


class TestDecomposition:
    def test_decomposition_is_performed(self, diffraction_pattern):
        s = Correlation2D(diffraction_pattern)
        s.decomposition()
        assert s.learning_results is not None

    def test_decomposition_class_assignment(self, diffraction_pattern):
        s = Correlation2D(diffraction_pattern)
        s.decomposition()
        assert isinstance(s, Correlation2D)



# -*- coding: utf-8 -*-
# Copyright 2016-2025 The pyXem developers
#
# This file is part of pyXem.
#
# pyXem is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyXem is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyXem.  If not, see <http://www.gnu.org/licenses/>.

"""VDF generator and associated tools."""
import numpy as np

import hyperspy.api as hs

from pyxem.signals.common_diffraction import OUT_SIGNAL_AXES_DOCSTRING
from pyxem.utils.virtual_images_utils import normalize_virtual_images, get_vectors_mesh
from pyxem.utils._deprecated import deprecated
from warnings import warn

NORMALISE_DOCSTRING = """normalize : boolean
            If True each VDF image is normalized so that the maximum intensity
            in each VDF is 1.
        """


class VirtualImageGenerator:
    """Generates virtual images for a specified signal and set of aperture.

    Attributes
    ----------
    signal : Diffraction2D or subclass
        The signal of electron diffraction patterns to be indexed.
    roi_list: list of roi
        The list of roi used to defined the integration area of the virtual
        images in the signal space.

    """

    def __init__(self, signal, *args, **kwargs):
        self.signal = signal
        self.roi_list = []
        warn(
            "The VirtualImageGenerator class is deprecated and will be removed in 1.0.0. "
            "Please use the pyxem.signals.Diffraction2D.get_virtual_image function instead.",
            DeprecationWarning,
        )

    @deprecated(
        alternative="pyxem.signals.Diffraction2D.get_azithumal_integral1d",
        since="0.17.0",
        removal="1.0.0",
    )
    def get_concentric_virtual_images(
        self, k_min, k_max, k_steps, normalize=False, out_signal_axes=None
    ):
        """
        Obtain the intensity scattered at each navigation position in an
        Diffraction2D Signal by summation over a series of concentric
        in annuli between a specified inner and outer radius in a number of
        steps.

        Parameters
        ----------
        k_min : float
            Minimum radius of the annular integration window in units of
            reciprocal space.

        k_max : float
            Maximum radius of the annular integration window in units of
            reciprocal space.

        k_steps : int
            Number of steps within the annular integration window
        %s
        %s

        Returns
        -------
        virtual_images : VirtualDarkFieldImage
            VirtualDarkFieldImage object containing virtual images for all
            steps within the annulus.
        """
        k_step = (k_max - k_min) / k_steps
        k0s = np.linspace(k_min, k_max - k_step, k_steps)
        k1s = np.linspace(k_min + k_step, k_max, k_steps)

        ks = np.array((k0s, k1s)).T

        roi_args_list = [(0, 0, r[1], r[0]) for r in ks]

        new_axis_dict = {
            "name": "Annular bins",
            "scale": k_step,
            "units": self.signal.axes_manager[-1].units,
            "offset": k_min,
        }

        return self._get_virtual_images(
            roi_args_list,
            normalize,
            new_axis_dict=new_axis_dict,
            out_signal_axes=out_signal_axes,
        )

    get_concentric_virtual_images.__doc__ %= (
        NORMALISE_DOCSTRING,
        OUT_SIGNAL_AXES_DOCSTRING,
    )

    def set_ROI_mesh(
        self,
        vector1_norm,
        vector2_norm,
        vector_norm_max,
        angle=0.0,
        shear=0.0,
        ROI_radius=None,
    ):
        """
        Set and display a mesh of ROI in the signal space. The ROIs are stored
        in the ``roi_list`` attribute and will be by the
        ``get_virtual_images_from_mesh`` method.

        Parameters
        ----------
        vector1_norm, vector2_norm : float
            The norm of one of the two vectors of the mesh. The position of the
            other vector is defined by the shear parameter.
        vector_norm_max : float
            The maximum value for the norm of each vector.
        angle : float, optional
            The rotation of the mesh in degree.
        shear : float, optional
            The shear of the mesh. It must be in the interval [0, 1].
            The default is 0.0.
        ROI_radius : None or float, optional
            The radius of the ROI. If None, half of the minimum value between
            ``vector1_norm`` and ``vector1_norm`` is used.
            The default is None.

        Returns
        -------
        None.

        """

        vectors = get_vectors_mesh(
            vector1_norm, vector2_norm, vector_norm_max, angle, shear
        )

        if len(self.roi_list) > 0:
            for roi in self.roi_list:
                roi.remove_widget(self.signal)
            self.roi_list = []

        if ROI_radius is None:
            ROI_radius = min(vector1_norm, vector2_norm) / 2

        roi_args_list = [(*v, ROI_radius) for v in vectors]

        if self.signal._plot is None:
            self.signal.plot()

        for roi_args in roi_args_list:
            roi = hs.roi.CircleROI(*roi_args)
            roi.add_widget(self.signal, axes=self.signal.axes_manager.signal_axes)
            self.roi_list.append(roi)

    def get_virtual_images_from_mesh(self, normalize=False, out_signal_axes=None):
        """
        Obtain the intensity scattered at each navigation position in an
        Diffraction2D Signal by summation over the ROIs defined in the
        ``roi_list`` attribute.

        Parameters
        ----------
        %s
        %s


        Returns
        -------
        virtual_images : VirtualDarkFieldImage
            VirtualDarkFieldImage object containing the virtual images for all
            ROIs, with the navigation axis as ROI index.

        """
        if len(self.roi_list) == 0:
            raise ValueError(
                "The `roi_list` attribute can't be of length 0. "
                "You can set it manually and use the convenience "
                "`set_ROI_mesh` method."
            )
        new_axis_dict = {"name": "ROI index"}
        out = self._get_virtual_images(
            self.roi_list,
            normalize,
            new_axis_dict=new_axis_dict,
            out_signal_axes=out_signal_axes,
        )

        return out

    get_virtual_images_from_mesh.__doc__ %= (
        NORMALISE_DOCSTRING,
        OUT_SIGNAL_AXES_DOCSTRING,
    )

    def _get_virtual_images(
        self, roi_list, normalize, new_axis_dict, out_signal_axes=None
    ):
        """
        Obtain the intensity scattered at each navigation position in an
        Diffraction2D Signal by summation over the roi defined by the
        ``roi_list`` parameter.

        Parameters
        ----------
        roi_list : list of hyperspy ROI or list of `hyperspy.roi.CircleROI` arguments
            List of ROI or Arguments required to initialise a CircleROI
        %s
        %s

        Returns
        -------
        virtual_images : VDFImage
            VDFImage object containing the virtual images
        """
        if isinstance(roi_list[0], hs.roi.CircleROI):
            self.roi_list = self.roi_list
        else:
            self.roi_list = [hs.roi.CircleROI(*r) for r in roi_list]

        vdfs = [
            self.signal.get_integrated_intensity(roi, out_signal_axes)
            for roi in self.roi_list
        ]

        vdfim = hs.stack(
            vdfs, new_axis_name=new_axis_dict["name"], show_progressbar=False
        )

        vdfim.set_signal_type("virtual_dark_field")

        if vdfim.metadata.has_item("Diffraction.integrated_range"):
            del vdfim.metadata.Diffraction.integrated_range
        vdfim.metadata.set_item(
            "Diffraction.roi_list", [f"{roi}" for roi in self.roi_list]
        )

        # Set new axis properties
        new_axis = vdfim.axes_manager[new_axis_dict["name"]]
        for k, v in new_axis_dict.items():
            setattr(new_axis, k, v)

        if normalize:
            vdfim.map(normalize_virtual_images, show_progressbar=False)

        return vdfim

    _get_virtual_images.__doc__ %= (NORMALISE_DOCSTRING, OUT_SIGNAL_AXES_DOCSTRING)


class VirtualDarkFieldGenerator(VirtualImageGenerator):
    """Generates VDF images for a specified signal and set of aperture
    positions.

    Attributes
    ----------
    signal : Diffraction2D of subclass
        The signal of diffraction patterns to be indexed.
    vectors: DiffractionVectors
        The vector positions, in calibrated units, at which to position
        integration windows for virtual dark field formation.

    """

    def __init__(self, signal, vectors, *args, **kwargs):
        super().__init__(signal)
        warn(
            "The VirtualDarkFieldGenerator class is deprecated and will be removed in 1.0.0. "
            "Please use the pyxem.signals.DiffractionVectors2D.to_roi function instead.",
            DeprecationWarning,
        )
        if len(vectors.axes_manager.signal_axes) == 0:
            unique_vectors = vectors.get_unique_vectors(*args, **kwargs)
        else:
            unique_vectors = vectors

        self.vectors = unique_vectors

    @deprecated(
        alternative="pyxem.signals.DiffractionVectors2D.to_roi",
        since="0.17.0",
        removal="1.0.0",
    )
    def get_virtual_dark_field_images(
        self, radius, normalize=False, out_signal_axes=None
    ):
        """Obtain the intensity scattered to each diffraction vector at each
        navigation position in an Diffraction2D Signal by summation in a
        circular window of specified radius.

        Parameters
        ----------
        radius : float
            Radius of the integration window - in units of the reciprocal
            space.
        %s
        %s

        Returns
        -------
        vdfs : VirtualDarkFieldImage
            VirtualDarkFieldImage object containing virtual dark field images
            for all unique vectors.
        """
        roi_args_list = [(v[1], v[0], radius, 0) for v in self.vectors.data]
        new_axis_dict = {"name": "Vector index"}
        vdfim = self._get_virtual_images(
            roi_args_list,
            normalize,
            new_axis_dict=new_axis_dict,
            out_signal_axes=out_signal_axes,
        )

        # Assign vectors used to generate images to vdfim attribute.
        vdfim.vectors = self.vectors

        return vdfim

    get_virtual_dark_field_images.__doc__ %= (
        NORMALISE_DOCSTRING,
        OUT_SIGNAL_AXES_DOCSTRING,
    )



# -*- coding: utf-8 -*-
# Copyright 2016-2025 The pyXem developers
#
# This file is part of pyXem.
#
# pyXem is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyXem is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyXem.  If not, see <http://www.gnu.org/licenses/>.

import pytest
import numpy as np

from pyxem.utils._correlations import (
    _correlation,
    _get_interpolation_matrix,
    _wrap_set_float,
)


class TestCorrelations:
    @pytest.fixture
    def ones_array(self):
        return np.ones((10, 20))

    @pytest.fixture
    def ones_zero(self):
        ones = np.ones((10, 20))
        ones[0:20:2, :] = 0
        return ones

    @pytest.fixture
    def ones_hundred(self):
        ones = np.ones((10, 20))
        ones[0:20:2, :] = 100
        return ones

    def test_correlation_ones(self, ones_array):
        c = _correlation(ones_array)
        np.testing.assert_array_equal(c, np.zeros((10, 20)))

    def test_correlations_axis(self, ones_zero):
        c = _correlation(ones_zero, axis=0, normalize=True)
        result = np.ones((10, 20))
        result[1::2, :] = -1
        np.testing.assert_array_equal(c, result)
        c = _correlation(ones_zero, axis=0, normalize=False)
        result = np.zeros((10, 20))
        result[0::2, :] = 5
        np.testing.assert_array_almost_equal(c, result)

    def test_wrap_set_float(self):
        target = np.zeros(20)
        assert 2 == np.sum(_wrap_set_float(target, bottom=10.0, top=15.5, value=2))
        target = np.zeros(20)
        assert 2 == np.sum(_wrap_set_float(target, bottom=10.0, top=10.5, value=2))
        target = np.zeros(20)
        assert 2 == np.sum(_wrap_set_float(target, bottom=10.0, top=10.0, value=2))

    def test_correlations_normalization(self, ones_hundred):
        c = _correlation(ones_hundred, axis=0, normalize=True)
        result = np.zeros((10, 20))
        result[1::2, :] = -0.96078816
        result[0::2, :] = 0.96078816
        np.testing.assert_array_almost_equal(c, result)
        c = _correlation(ones_hundred, axis=1, normalize=True)
        result = np.zeros((10, 20))
        np.testing.assert_array_almost_equal(c, result)

    def test_correlations_mask(self, ones_hundred):
        m = np.zeros((10, 20))
        m[2:4, :] = 1
        c = _correlation(ones_hundred, axis=0, normalize=True, mask=m)
        print(c)
        result = np.zeros((10, 20))
        result[1::2, :] = -0.96078816
        result[0::2, :] = 0.96078816
        np.testing.assert_array_almost_equal(c, result)

    def test_correlations_wrapping(
        self, ones_hundred
    ):  # Need to do extra checks to assure this is correct
        m = np.zeros((10, 20))
        m[2:4, :] = 1
        c = _correlation(ones_hundred, axis=0, normalize=True, wrap=False)
        print(c)
        result = np.zeros((10, 20))
        result[0::2, :] = 2.26087665
        result[1::2, :] = -0.93478899
        np.testing.assert_array_almost_equal(c, result)

    @pytest.mark.parametrize(
        "angles",
        np.array(
            [
                [0.1],
                [
                    0.25,
                ],
                [
                    0.5,
                ],
                [
                    0.9,
                ],
            ]
        ),
    )
    @pytest.mark.parametrize("angular_range", [0, 4 * np.pi / 90])
    def test_get_interpolation_matrix(self, angles, angular_range):
        m = np.zeros((10, 20))
        m[2:4, :] = 1
        c = _get_interpolation_matrix(
            angles, num_points=90, angular_range=angular_range
        )
        com = np.average(range(90), weights=c)
        np.testing.assert_almost_equal(com, 90 * angles[0])



# Copyright 2016-2025 The pyXem developers
#
# This file is part of pyXem.
#
# pyXem is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyXem is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyXem.  If not, see <http://www.gnu.org/licenses/>.

from hyperspy._signals.lazy import LazySignal

from pyxem.signals.diffraction_vectors import DiffractionVectors
from pyxem.utils.vectors import (
    to_cart_three_angles,
    polar_to_cartesian,
    get_three_angles,
)


class PolarVectors(DiffractionVectors):
    """
    Class for representing polar vectors in reciprocal space. The first
    two columns are the polar coordinates of the vectors, (r, theta).
    """

    _signal_dimension = 0
    _signal_type = "polar_vectors"

    def get_angles(
        self,
        intensity_index=2,
        intensity_threshold=None,
        accept_threshold=0.05,
        min_k=0.05,
        min_angle=None,
        **kwargs
    ):
        """Calculate the angles between pairs of 3 diffraction vectors.

        Parameters
        ----------
        intensity_index : int, optional
            The column index of the intensity column.  The default is 2.
        intensity_threshold : float, optional
            The minimum intensity for a vector to be considered.  The default is None.
        accept_threshold : float, optional
            The maximum difference between the inscribed angle and the
            reduced angle for the inscribed angle to be accepted.  The default is 0.05.
        min_k : float, optional
            The minimum average delta k between the three vectors
             for a angle to be considered.  The default is 0.05.
        min_angle : float, optional
            The minimum inscribed angle for an angle to be considered.  The default is None.
            Which means that two vectors even with a small angle between them will be
            considered.
        **kwargs:
            Keyword arguments to :meth:`hyperspy.api.signals.BaseSignal.map`.

        Returns
        -------
        angles : BaseSignal
            A signal with navigation dimensions as the original diffraction
            vectors containg an array of inscribed angles at each
            navigation position.

        """
        angles = self.map(
            get_three_angles,
            k_index=0,
            angle_index=1,
            intensity_index=intensity_index,
            intensity_threshold=intensity_threshold,
            accept_threshold=accept_threshold,
            min_k=min_k,
            min_angle=min_angle,
            inplace=False,
            ragged=True,
            silence_warnings=True,
            **kwargs,
        )
        # set the column names
        col_names = ["k", "delta phi", "min-angle", "intensity", "reduced-angle"]

        # Maybe we should pass any other vector columns through?
        angles.column_names = col_names
        angles.units = ["nm^-1", "rad", "rad", "a.u.", "rad"]
        return angles

    @property
    def has_delta_phi(self):
        """If the polar vector has the delta phi column.

        If True, the vectors are in the form (r, delta phi, min-angle) as a result of the
         ``get_angles method``.  If False, the vectors are in the form (r, theta)."""
        return self.column_names[1] == "delta phi"

    def to_cartesian(self, has_delta_angle=None):
        """Convert the vectors to cartesian coordinates.

        Parameters
        ----------
        has_delta_angle : bool
            If True, the vectors are in the form (r, delta phi, min-angle).  If False,
            the vectors are in the form (r, theta).

        Returns
        -------
        DiffractionVectors
            The vectors in cartesian coordinates.
        """
        if has_delta_angle is None:
            has_delta_angle = self.has_delta_phi
        if has_delta_angle:
            cartesian_vectors = self.map(
                to_cart_three_angles,
                silence_warnings=True,
                inplace=False,
                ragged=True,
            )
        else:
            cartesian_vectors = self.map(
                polar_to_cartesian,
                inplace=False,
                ragged=True,
                silence_warnings=True,
            )
        cartesian_vectors.column_names = ["x", "y"]
        cartesian_vectors.units = [self.units[0], self.units[0]]
        cartesian_vectors.set_signal_type("diffraction_vectors")
        return cartesian_vectors

    def to_markers(self, has_delta_angle=None, cartesian=True, **kwargs):
        """Convert the vectors to markers to be plotted on a signal.

        Parameters
        ----------
        has_delta_angle : bool, optional
            If the vectors are polar in the form (r, theta), then this parameter should
            be set to False.  If the vectors are in the form (r, delta phi, min-angle), then
            this parameter should be set to True.  The default is None which will infer
            the format from the column names.
        cartesian : bool, optional
            If True, the vectors will be converted to cartesian coordinates before plotting.
            The default is True.
        **kwargs :
            Keyword arguments to be passed to the :class:`hyperspy.api.plot.markers.Point` class.

        Returns
        -------
        :class:`hyperspy.api.plot.markers.Point`
            A Point object containing the markers.
        """
        if cartesian:
            vectors = self.to_cartesian(has_delta_angle=has_delta_angle)
        else:
            vectors = self
        return vectors.to_markers(**kwargs)


class LazyPolarVectors(LazySignal, PolarVectors):
    pass



# -*- coding: utf-8 -*-
# Copyright 2016-2025 The pyXem developers
#
# This file is part of pyXem.
#
# pyXem is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyXem is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyXem.  If not, see <http://www.gnu.org/licenses/>.

import numpy as np

from pyxem.signals import Diffraction2D


class TestAdjustMetadata:
    def test_add_navigation_signal(self):
        d = Diffraction2D(np.ones((10, 10, 10, 10)))
        data = np.ones((10, 10))
        d.add_navigation_signal(data=data, name="Test", unit="nm", nav_plot=True)
        d.add_navigation_signal(data=data, name="Test2", unit="nm", nav_plot=True)
        np.testing.assert_equal(d.metadata.Navigation_signals.Test["data"], data)
        np.testing.assert_equal(d.metadata.Navigation_signals.Test2["data"], data)



# -*- coding: utf-8 -*-
# Copyright 2016-2025 The pyXem developers
#
# This file is part of pyXem.
#
# pyXem is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyXem is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyXem.  If not, see <http://www.gnu.org/licenses/>.

import pytest
import numpy as np

from pyxem.signals.pair_distribution_function1d import PairDistributionFunction1D


def test_generate_signal():
    data = np.ones((1, 10)) * np.arange(1, 5).reshape(4, 1)
    data = data.reshape(2, 2, 10)
    pdf = PairDistributionFunction1D(data)
    assert isinstance(pdf, PairDistributionFunction1D)
    pdf.normalise_signal()
    assert np.equal(pdf, np.ones((2, 2, 10)))



# -*- coding: utf-8 -*-
# Copyright 2016-2025 The pyXem developers
#
# This file is part of pyXem.
#
# pyXem is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyXem is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyXem.  If not, see <http://www.gnu.org/licenses/>.


import logging

try:
    import cupy as cp

    CUPY_INSTALLED = True
except ImportError:
    CUPY_INSTALLED = False
from pyxem import components
from pyxem import signals
from pyxem import generators
from pyxem import data

from . import release_info


_logger = logging.getLogger(__name__)


__all__ = [
    "components",
    "generators",
    "signals",
    "data",
    "CUPY_INSTALLED",
]

__version__ = release_info.version
__author__ = release_info.author
__copyright__ = release_info.copyright
__credits__ = release_info.credits
__license__ = release_info.license
__maintainer__ = release_info.maintainer
__email__ = release_info.email
__status__ = release_info.status



# -*- coding: utf-8 -*-
# Copyright 2016-2025 The pyXem developers
#
# This file is part of pyXem.
#
# pyXem is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyXem is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyXem.  If not, see <http://www.gnu.org/licenses/>.

import pytest
import numpy as np
import dask.array as da

from hyperspy.signals import Signal2D, Signal1D

from pyxem.signals import (
    PolarDiffraction2D,
    LazyPolarDiffraction2D,
    Correlation2D,
    Power2D,
    Correlation1D,
)


class TestComputeAndAsLazy2D:
    def test_2d_data_compute(self):
        dask_array = da.random.random((100, 150), chunks=(50, 50))
        s = LazyPolarDiffraction2D(dask_array)
        scale0, scale1, metadata_string = 0.5, 1.5, "test"
        s.axes_manager[0].scale = scale0
        s.axes_manager[1].scale = scale1
        s.metadata.Test = metadata_string
        s.compute()
        assert s.__class__ == PolarDiffraction2D
        assert not hasattr(s.data, "compute")
        assert s.axes_manager[0].scale == scale0
        assert s.axes_manager[1].scale == scale1
        assert s.metadata.Test == metadata_string
        assert dask_array.shape == s.data.shape

    def test_4d_data_compute(self):
        dask_array = da.random.random((4, 4, 10, 15), chunks=(1, 1, 10, 15))
        s = LazyPolarDiffraction2D(dask_array)
        s.compute()
        assert s.__class__ == PolarDiffraction2D
        assert dask_array.shape == s.data.shape

    def test_2d_data_as_lazy(self):
        data = np.random.random((100, 150))
        s = PolarDiffraction2D(data)
        scale0, scale1, metadata_string = 0.5, 1.5, "test"
        s.axes_manager[0].scale = scale0
        s.axes_manager[1].scale = scale1
        s.metadata.Test = metadata_string
        s_lazy = s.as_lazy()
        assert s_lazy.__class__ == LazyPolarDiffraction2D
        assert hasattr(s_lazy.data, "compute")
        assert s_lazy.axes_manager[0].scale == scale0
        assert s_lazy.axes_manager[1].scale == scale1
        assert s_lazy.metadata.Test == metadata_string
        assert data.shape == s_lazy.data.shape

    def test_4d_data_as_lazy(self):
        data = np.random.random((4, 10, 15))
        s = PolarDiffraction2D(data)
        s_lazy = s.as_lazy()
        assert s_lazy.__class__ == LazyPolarDiffraction2D
        assert data.shape == s_lazy.data.shape


class TestCorrelations:
    @pytest.fixture
    def flat_pattern(self):
        pd = PolarDiffraction2D(data=np.ones(shape=(2, 2, 5, 5)))
        pd.axes_manager.signal_axes[0].scale = 0.5
        pd.axes_manager.signal_axes[0].name = "theta"
        pd.axes_manager.signal_axes[1].scale = 2
        pd.axes_manager.signal_axes[1].name = "k"
        return pd

    def test_correlation_signal(self, flat_pattern):
        ac = flat_pattern.get_angular_correlation()
        assert isinstance(ac, Correlation2D)

    def test_axes_transfer(self, flat_pattern):
        ac = flat_pattern.get_angular_correlation()
        assert (
            ac.axes_manager.signal_axes[0].scale
            == flat_pattern.axes_manager.signal_axes[0].scale
        )
        assert (
            ac.axes_manager.signal_axes[1].scale
            == flat_pattern.axes_manager.signal_axes[1].scale
        )
        assert (
            ac.axes_manager.signal_axes[1].name
            == flat_pattern.axes_manager.signal_axes[1].name
        )

    @pytest.mark.parametrize(
        "mask", [None, np.zeros(shape=(5, 5)), Signal2D(np.zeros(shape=(2, 2, 5, 5)))]
    )
    def test_masking_correlation(self, flat_pattern, mask):
        ap = flat_pattern.get_angular_correlation(mask=mask)
        assert isinstance(ap, Correlation2D)

    def test_correlation_inplace(self, flat_pattern):
        ac = flat_pattern.get_angular_correlation(inplace=True)
        assert ac is None
        assert isinstance(flat_pattern, Correlation2D)

    @pytest.mark.parametrize(
        "mask", [None, np.zeros(shape=(5, 5)), Signal2D(np.zeros(shape=(2, 2, 5, 5)))]
    )
    def test_masking_angular_power(self, flat_pattern, mask):
        ap = flat_pattern.get_angular_power(mask=mask)
        print(ap)
        assert isinstance(ap, Power2D)

    def test_axes_transfer_power(self, flat_pattern):
        ac = flat_pattern.get_angular_power()
        assert ac.axes_manager.signal_axes[0].scale == 1
        assert (
            ac.axes_manager.signal_axes[1].scale
            == flat_pattern.axes_manager.signal_axes[1].scale
        )
        assert (
            ac.axes_manager.signal_axes[1].name
            == flat_pattern.axes_manager.signal_axes[1].name
        )

    def test_power_inplace(self, flat_pattern):
        ac = flat_pattern.get_angular_power(inplace=True)
        assert ac is None
        assert isinstance(flat_pattern, Power2D)


class TestPearsonCorrelation:
    @pytest.fixture
    def flat_pattern(self):
        rng = np.random.default_rng(seed=1)
        pd = PolarDiffraction2D(data=rng.random((2, 2, 15, 50)))
        pd.axes_manager.signal_axes[0].scale = 0.5
        pd.axes_manager.signal_axes[0].name = "theta"
        pd.axes_manager.signal_axes[1].scale = 0.1
        pd.axes_manager.signal_axes[1].name = "k"
        return pd

    @pytest.mark.parametrize("krange", [None, (0, 4), (0.5, 1.4)])
    def test_pearson_correlation_signal(self, flat_pattern, krange):
        rho = flat_pattern.get_full_pearson_correlation(krange=krange)
        assert isinstance(rho, Signal1D)
        rhok = flat_pattern.get_resolved_pearson_correlation(krange=krange)
        assert isinstance(rhok, Signal2D)

    @pytest.mark.parametrize("inplace", (True, False))
    @pytest.mark.parametrize("krange", [None, (0, 10), (0.5, 1.4)])
    def test_full_pearson_correlation_results(self, flat_pattern, krange, inplace):
        out = flat_pattern.get_full_pearson_correlation(
            krange=krange,
            inplace=inplace,
        )
        if inplace:
            assert out is None
            out = flat_pattern
        else:
            # check the original signal is not changed
            assert flat_pattern.axes_manager[-1].size == 15

        assert isinstance(out, Correlation1D)
        np.testing.assert_allclose(np.zeros((2, 2, 49)), out.data[..., 1:], atol=0.2)

    @pytest.mark.parametrize("inplace", (True, False))
    @pytest.mark.parametrize("krange", [None, (0, 10), (0.5, 1.4)])
    def test_resolved_pearson_correlation_results(self, flat_pattern, krange, inplace):
        out = flat_pattern.get_resolved_pearson_correlation(
            krange=krange,
            inplace=inplace,
        )
        if inplace:
            assert out is None
            out = flat_pattern
        else:
            # check the original signal is not changed
            assert flat_pattern.axes_manager[-1].size == 15
            assert flat_pattern.axes_manager[-2].size == 50

        assert isinstance(out, Correlation2D)
        np.testing.assert_allclose(
            np.zeros((2, 2, 49)), np.mean(out.data[..., 1:], axis=-2), atol=0.2
        )

    def test_full_pearson_correlation_inplace(self, flat_pattern):
        rho = flat_pattern.get_full_pearson_correlation(inplace=True)
        assert rho is None
        assert isinstance(flat_pattern, Correlation1D)

    def test_resolved_pearson_correlation_inplace(self, flat_pattern):
        rho = flat_pattern.get_resolved_pearson_correlation(inplace=True)
        assert rho is None
        assert isinstance(flat_pattern, Correlation2D)

    def test_axes_transfer(self, flat_pattern):
        rho = flat_pattern.get_full_pearson_correlation()
        assert (
            rho.axes_manager.signal_axes[0].scale
            == flat_pattern.axes_manager.signal_axes[0].scale
        )

        rhok = flat_pattern.get_resolved_pearson_correlation()
        assert (
            rhok.axes_manager.signal_axes[0].scale
            == flat_pattern.axes_manager.signal_axes[0].scale
        )
        assert (
            rhok.axes_manager.signal_axes[1].scale
            == flat_pattern.axes_manager.signal_axes[1].scale
        )

    @pytest.mark.parametrize("mask", [None, np.zeros(shape=(15, 50))])
    def test_masking_pearson_correlation(self, flat_pattern, mask):
        rho_0 = flat_pattern.get_full_pearson_correlation(mask=mask)
        assert isinstance(rho_0, Correlation1D)
        rho = flat_pattern.get_full_pearson_correlation(mask=mask, krange=(0, 4))
        assert isinstance(rho, Correlation1D)

        rhok_0 = flat_pattern.get_resolved_pearson_correlation(mask=mask)
        assert isinstance(rhok_0, Correlation2D)
        rhok = flat_pattern.get_resolved_pearson_correlation(mask=mask, krange=(0, 4))
        assert isinstance(rhok, Correlation2D)


class TestDecomposition:
    def test_decomposition_is_performed(self, diffraction_pattern):
        s = PolarDiffraction2D(diffraction_pattern)
        s.decomposition()
        assert s.learning_results is not None

    def test_decomposition_class_assignment(self, diffraction_pattern):
        s = PolarDiffraction2D(diffraction_pattern)
        s.decomposition()
        assert isinstance(s, PolarDiffraction2D)


class TestSubtractingDiffractionBackground:
    @pytest.fixture
    def noisy_data(self):
        data = np.random.rand(3, 2, 20, 15)
        data[:, :, 10:12, 7:9] = 100
        dp = PolarDiffraction2D(data)
        return dp

    @pytest.mark.parametrize(
        ["method", "kwargs"],
        [
            ("radial median", {}),
            ("radial percentile", {"percentile": 40}),
            pytest.param(
                "this method does not exist",
                {},
                marks=pytest.mark.xfail(raises=NotImplementedError),
            ),
        ],
    )
    def test_subtract_backgrounds(self, method, kwargs, noisy_data):
        kwargs["inplace"] = False

        subtracted = noisy_data.subtract_diffraction_background(method=method, **kwargs)
        assert isinstance(subtracted, PolarDiffraction2D)
        assert subtracted.data.shape == noisy_data.data.shape



import numpy as np
from pyxem.data.dummy_data import make_diffraction_test_data as mdtd
import hyperspy.api as hs


def tilt_boundary_data(correct_pivot_point=True):
    di = mdtd.DiffractionTestImage(intensity_noise=False)
    di.add_disk(x=128, y=128, intensity=10.0)  # Add a zero beam disk at the center
    di.add_cubic_disks(vx=20, vy=20, intensity=2.0, n=5)
    di.add_background_lorentz()
    di_rot = di.copy()
    di_rot.rotation = 10
    dtd = mdtd.DiffractionTestDataset(10, 10, 256, 256)
    position_array = np.ones((10, 10), dtype=bool)
    position_array[:5] = False
    dtd.add_diffraction_image(di, position_array)
    dtd.add_diffraction_image(di_rot, np.invert(position_array))
    s = dtd.get_signal()
    if not correct_pivot_point:
        # Shifting the zero beam away from the center
        xx, yy = np.meshgrid(range(10), range(10))
        shifts = np.stack([xx * 0.5, yy * 0.5], axis=-1)
        s.center_direct_beam(shifts=hs.signals.Signal1D(shifts))
    s.axes_manager.signal_axes.set(scale=0.3, offset=-0.3 * 128)
    return s
